#!amake
#
# Makefile to compile sswf (Script to SWF) with amake
# Written by Alexis Wilke for Made to Order Software, Ltd. (c) 2002-2005
#
# Copyright (c) 2002-2005 Made to Order Software, Ltd.
#
# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and
# associated documentation files (the "Software"), to
# deal in the Software without restriction, including
# without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the
# following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial
# portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
# ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
# EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# Get the version from the libsswf-version.h file instead (we thus avoid
# data duplication)
#VERSION=x.y.z
# The following ensures we can extract the version, it's not 100%
# correct since if a 2nd number with the same form appears in the
# version file then it will be returned too
LIBSSWF_VERSION<=include/sswf/libsswf-version.h
VERSION:=$(LIBSSWF_VERSION::[0-9][0-9]*\.[0-9][0-9]*\.[0-9][0-9]*)


# Let's ensure that all the necessary directories
# exist from the start; this makes amake happier;
# in MS-Windows 2000 you can test if a file exists
# I'm not sure that's valid under older versions
#for p in modules bin lib VC \
		tmp\backup\src\tools tmp\backup\src\sswf tmp\backup\src\lib \
		tmp\object\src\tools tmp\object\src\sswf tmp\object\src\lib
#shell if not exist $p mkdir $p
#endfor

# For cl we need to copy some header files...
#for p in include\jpeglib.h include\jerror.h include\jmorecfg.h include\zconf.h \
		include\zlib.h include\iconv.h
#shell if not exist $p copy ..\..\..\..\$p include
#endfor

#shell sed -e "s/#if 1/#if 0/" include\zconf.h >tmp\zconf.h
#shell copy tmp\zconf.h include

CINCLUDEPATH:=include;src/sswf
CSUFFIXES:=.c .c++ .cpp .cxx .h .h++ .hpp .s .S .l .y

CC:=cl	#gcc
CXX:=cl	#g++
LD:=ld
#CFLAGS:=-O2 -ggdb3 -I/AllDev/MinGW/3.3/include/freetype2 -Wall -Werror -DDEBUG=1 -DYYDEBUG=1 -Iinclude
CFLAGS:=/O2 /GR /I/AllDev/MinGW/3.3/include/freetype2 /DDEBUG=1 /DYYDEBUG=1 /D_MSVC=1 /DBYTE_ORDER=1 /DLITTLE_ENDIAN=1 /DBIG_ENDIAN=0 /Iinclude
CXXFLAGS:=$(CFLAGS)
COPY:=cp -f
RENAME:=rename
RM:=rm -r
RMDIR:=rmdir /s /q
MKDIR:=mkdir
AR:=ar
YFLAGS:=-t		# enable -t for debug
BISON:=bison

LIBSRC:=$(wildcard src\lib\libsswf_*.cpp)
RAWSSWFSRC:=$(wildcard src\sswf\*.[cy]*)
SSWFSRC:=$(RAWSSWFSRC:-*~)	# no backup (some people will use ~ at the end of filenames)
TOOLSSRC:=$(wildcard src\tools\*.c)
MISCSRC:=$(wildcard src\misc\*.c)

LIBOBJ:=$(LIBSRC:*.cpp=tmp\\object\\*.obj:U)
SSWFOBJ:=$(SSWFSRC:-*~:*.*=tmp\\object\\*.obj:U)
TOOLSOBJ:=$(TOOLSSRC:*.*=tmp\\object\\*.obj:U)
MISCOBJ:=$(MISCSRC:*.*=tmp\\object\\*.obj:U)

SAMPLES:=new-anim dyn-button scroller web-site-anim



EXE:=bin\\swf_dump.exe bin\\sswf.exe bin\\ft2sswf.exe
all: $(EXE) $(EXE:*.exe=*-static.exe) lib\libsswf.a .PHONY
	@echo "Done!"


bin\\swf_dump.exe: $(.TARGET:bin\\*.exe=tmp\\object\\src\\tools\\*.obj)
	$(CC) $(CFLAGS) /Fo$(.TARGET) $(.FIRSTSRC) -lz

bin\\swf_dump-static.exe: $(.TARGET:bin\\*-static.exe=tmp\\object\\src\\tools\\*.obj)
	$(CC) -static $(CFLAGS) /Fo$(.TARGET) $(.FIRSTSRC) -lz


bin\\ft2sswf.exe: $(.TARGET:bin\\*.exe=tmp\\object\\src\\tools\\*.obj)
	$(CC) $(CFLAGS) /Fo$(.TARGET) $(.FIRSTSRC) -lfreetype

bin\\ft2sswf-static.exe: $(.TARGET:bin\\*-static.exe=tmp\\object\\src\\tools\\*.obj)
	$(CC) -static $(CFLAGS) /Fo$(.TARGET) $(.FIRSTSRC) -lfreetype -lz -lgw32c


bin\\sswf.exe: $(.TARGET:bin\\*.exe=tmp\\object\\src\\sswf\\*.obj) $(SSWFOBJ) lib\\libsswf.a
	$(CXX) $(CFLAGS) /Fo$(.TARGET) $(SSWFOBJ) lib/libsswf.a -lm -lz -ljpeg -liconv

bin\\sswf-static.exe: $(.TARGET:bin\\*-static.exe=tmp\\object\\src\\sswf\\*.obj) $(SSWFOBJ) lib\\libsswf.a
	$(CXX) -static $(CFLAGS) /Fo$(.TARGET) $(SSWFOBJ) lib/libsswf.a ../../../../lib/libiconv.lib -lm -lz -ljpeg






lib/libsswf.a: $(LIBOBJ)
	$(AR) cr $(.TARGET) $(LIBOBJ)


ALLSRC:=$(LIBSRC) $(SSWFSRC) $(TOOLSSRC) $(MISCSRC)
ALLCXXOBJ:=$(ALLSRC::*.cpp=tmp\\object\\*.obj)
ALLCOBJ:=$(ALLSRC::*.c=tmp\\object\\*.obj)
ALLYACCOBJ:=$(ALLSRC::*.y=tmp\\object\\*.obj)
ALLLEXOBJ:=$(ALLSRC::*.l=tmp\\object\\*.obj)



$(ALLCXXOBJ): $(.TARGET:tmp\\object\\*.obj=*.cpp:I:src\\sswf\\sswf_grammar.h=tmp\\object\\src\\sswf\\sswf_grammar.h)
	$(CXX) /c $(CXXFLAGS) /I$(.TARGET:tmp\\object\\*\\*\\*.obj=tmp\\object\\*\\*) /Fo$(.TARGET) $(.FIRSTSRC)
	@$(COPY) "$(.FIRSTSRC)" "tmp\\backup\\$(.FIRSTSRC)"

$(ALLCOBJ): $(.TARGET:tmp\\object\\*.obj=*.c:I)
	$(CC) -c $(CFLAGS) -o $(.TARGET) $(.FIRSTSRC)
	@$(COPY) "$(.FIRSTSRC)" "tmp\\backup\\$(.FIRSTSRC)"


tmp/object/src/sswf/%.o: tmp/object/src/sswf/%.c
	$(CC) -c $(CFLAGS) -Isrc/sswf -o $(.TARGET) $(.IMPSRC)

tmp/object/src/sswf/%.c tmp/object/src/sswf/%.h: src/sswf/%.y $(.TARGET:tmp/object/*.c=*.y:-I) .PRECIOUS
	$(BISON) -v -d $(YFLAGS) -b $(.TARGET:R) $(.IMPSRC) -o $(.TARGET:*.h=*.c)
	@$(COPY) "$(.IMPSRC)" "tmp\\backup\\$(.IMPSRC)"


#for sample in $(SAMPLES)
$(sample): sswf.exe
	cd samples\$(sample) & run
#endfor

samples: $(SAMPLES)
	@echo Samples built successfully


TOPACK_BIN:=bin\sswf.exe bin\swf_dump.exe bin\ft2sswf.exe
TOPACK_LIB:=lib\libsswf.a include\sswf\libsswf.h include\sswf\libsswf-version.h include\sswf\libsswf-config.h
TOPACK_GRLX:=VC\sswf_grammar.h VC\sswf_grammar.c
TOPACK_ASIS:=VC\sswf_grammar.h VC\sswf_grammar.c $(SAMPLES:*=samples/*/run.bat)

INCLUDES:=$(wildcard */*/*.h)
TOPACK_SRC:=$(LIBSRC) $(SSWFSRC) $(INCLUDES) src\tools\ft2sswf.c src\tools\swf_dump.c \
		misc\sswf.vim misc\xswf.vim doc\CHANGES.txt dosamakefile
		#ged.bat

MOD_PATH:=modules

DST:=arch
DST_DLL:=$(DST)\\mingw
DST_STATIC:=$(DST)\\win32
DST_SRC:=$(DST)\\mingw-3.3

new_grammar: tmp\object\src\sswf\sswf_grammar.h tmp\object\src\sswf\sswf_grammar.c .PHONY
	$(COPY) tmp\object\src\sswf\sswf_grammar.h VC\sswf_grammar.h
	$(COPY) tmp\object\src\sswf\sswf_grammar.c VC\sswf_grammar.c


tar packages: new_grammar $(TOPACK_BIN) $(TOPACK_BIN:*.exe=*-static.exe) $(TOPACK_LIB) .PHONY
	-$(RMDIR) $(DST) 2>ignore.tmp
	REM
	REM Create the archive with the dynamically compiled binaries (need .dll)
	REM
	$(MKDIR) $(DST_DLL)\bin
	$(MKDIR) $(DST_DLL)\lib
	$(MKDIR) $(DST_DLL)\include\sswf
	strip $(TOPACK_BIN)
	$(subst ",,$(TOPACK_BIN:*.exe=$(COPY) *.exe $(DST_DLL)\\[1]*.exe &))
	$(subst ",,$(TOPACK_LIB:*.*=$(COPY) *.* $(DST_DLL)\\[1]*.* &))
	tar cf tmp.tar $(DST_DLL:*\\\\*=*/*)
	bzip2 -c9 tmp.tar >$(MOD_PATH)\$(DST_DLL:arch\\*=*)-$(VERSION)-bin.tar.bz2
	REM
	REM Create the archive with the static compiled binaries (don't need the .dll's)
	REM
	$(MKDIR) $(DST_STATIC)\bin
	$(MKDIR) $(DST_STATIC)\lib
	$(MKDIR) $(DST_STATIC)\include\sswf
	$(MKDIR) $(DST_STATIC)\VC
	strip $(TOPACK_BIN:*.exe=*-static.exe)
	$(subst ",,$(TOPACK_BIN:*.exe=$(COPY) *-static.exe $(DST_STATIC)\\[1]*.exe &))
	$(subst ",,$(TOPACK_LIB:*.*=$(COPY) *.* $(DST_STATIC)\\[1]*.* &))
	$(subst ",,$(TOPACK_GRLX:*.*=$(COPY) *.* $(DST_STATIC)\\[1]*.* &))
	tar cf tmp.tar $(DST_STATIC:*\\\\*=*/*)
	bzip2 -c9 tmp.tar >$(MOD_PATH)\$(DST_STATIC:arch\\*=*)-$(VERSION)-bin.tar.bz2
	REM
	REM Create an archive with the source so we get the changes
	REM
	$(MKDIR) $(DST_SRC)\src\lib
	$(MKDIR) $(DST_SRC)\src\sswf
	$(MKDIR) $(DST_SRC)\src\tools
	$(MKDIR) $(DST_SRC)\include\sswf
	$(MKDIR) $(DST_SRC)\doc
	$(MKDIR) $(DST_SRC)\misc
	$(subst ",,$(TOPACK_SRC:*=$(COPY) * $(DST_SRC)\\[1]* &))
	tar cf tmp.tar $(DST_SRC:*\\\\*=*/*) $(TOPACK_ASIS:*\\*\\*=*/*/*)
	bzip2 -c9 tmp.tar >$(MOD_PATH)\$(DST_SRC:arch\\*=*)_sswf-$(VERSION)-src.tar.bz2
	REM
	REM Cleanup
	REM
	-$(RMDIR) $(DST)
	-$(RM) tmp.tar

LFWGE=$(wildcard *.cpp)
to_vc:
	$(subst ",,$(LFWGE:*.cpp=$(RENAME) *.cpp [1]*.cpp &))

LFWBE=$(wildcard *.cpp)
to_gcc:
	$(subst ",,$(LFWBE:*.cpp=$(RENAME) *.cpp [1]*.cpp &)



clobber: rmtarget .PHONY
	-$(RMDIR) tmp 2>ignore.tmp
	-$(RM) .amakeautodepend
	@echo "Output packages not removed with amake clobber"

clean cleaup: rmtarget .PHONY
	-$(RMDIR) tmp/object 2>ignore.tmp

rmtarget: .PHONY
	-$(RMDIR) lib bin 2>ignore.tmp
	-$(RM) core 2>ignore.tmp


