/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// Action
Action::Action(TagBase *tag, action_t action)
	: f_action(action), f_tag(tag)
{
	static int		action_to_version[256] = {
		/* 0x00-0x0F */
		 1,  0,  0,  0,  1,  1,  1,  1,  1,  1, -4,  4,  4,  4, -4, -4,
		/* 0x10-0x1F */
		 4,  4,  4, -4,  4,  4,  0,  4,  4,  0,  0,  0,  4,  4,  0,  0,
		/* 0x20-0x2F */
		 4,  4,  4,  4,  4,  4,  4,  4,  4, -4,  7,  7,  7,  0,  0,  0,
		/* 0x30-0x3F */
		 4,  4,  4,  4,  4,  4,  4,  0,  0,  0,  5,  5,  5,  5,  5,  5,
		/* 0x40-0x4F */
		 5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,
		/* 0x50-0x5F */
		 5,  5,  5,  5,  6,  6,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
		/* 0x60-0x6F */
		 5,  5,  5,  5,  5,  5,  6,  6,  6,  7,  0,  0,  0,  0,  0,  0,
		/* 0x70-0x7F */
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
		/* 0x80-0x8F */
		 0,  1,  0,  1,  0,  0,  0,  5,  5,  0,  1,  1,  3,  4,  7,  7,
		/* 0x90-0x9F */
		/*   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F */
		 0,  0,  0,  0,  5,  0,  4,  0,  0,  4,  4,  5,  0,  4,  4,  4,
		/* 0xA0-0xAF */
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
		/* 0xB0-0xBF */
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
		/* 0xC0-0xCF */
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
		/* 0xD0-0xDF */
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
		/* 0xE0-0xEF */
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
		/* 0xF0-0xFF */
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
	};

	assert(f_tag != 0, "all actions need to have a tag attached to them");

	if(action == ACTION_LABEL) {
		f_min_version = 1;
	}
	else {
		assert(action >= ACTION_min && action <= ACTION_max, "trying to save an invalid action number");

		/* I'm not too sure why I wanted the -4, I think that's because these
		 * instructions are supposedly different in V4.x and V5.x plugins but
		 * that doesn't seem to really be the case. Anyway, we ignore it with
		 * the use of the labs() call below.
		 */
		f_min_version = (unsigned char) labs(action_to_version[action]);

		assert(f_min_version != 0, "unknown action #%d or undefined version -- please define the version in Action::Action()", action);
	}
}


#if DEBUG
Action::Action(const Action& action)
	 : f_action(action.f_action)	// required by Visual C++ because f_action is constant
{
}
Action& Action::operator = (const Action& action)
{
	return *this;
}
#endif


int Action::GetMaximumRegister(const Vectors& list)
{
	Action		*a;
	int		idx, max, r, register_number;
	Vectors		*sub_list;

	register_number = -1;
	max = list.Count();
	for(idx = 0; idx < max; idx++) {
		a = dynamic_cast<Action *>(list.Get(idx));
		if(a->f_action >= 128) {
			sub_list = a->SubList();
			if(sub_list != 0) {
				// Function, With and others are nested type
				// of entries and they may include higher
				// action versions
				r = GetMaximumRegister(*sub_list);
				if(r > register_number) {
					register_number = r;
				}
			}
		}
		r = a->GetMaxRegister();
		if(r > register_number) {
			register_number = r;
		}
	}

	return register_number;
}


int Action::MinimumListVersion(const Vectors& list)
{
	Action		*a;
	int		idx, max, v, version;
	Vectors		*sub_list;

	version = 1;
	max = list.Count();
	for(idx = 0; idx < max; idx++) {
		a = dynamic_cast<Action *>(list.Get(idx));
		if(a->f_action >= 128) {
			sub_list = a->SubList();
			if(sub_list != 0) {
				// Function, With and others are nested type
				// of entries and they may include higher
				// action versions
				v = MinimumListVersion(*sub_list);
				if(v > version) {
					version = v;
				}
			}
		}
		// Now Version() is a virtual function since
		// ActionTry has to check three sub-lists
		v = a->Version();
		if(v > version) {
			version = v;
		}
	}

	return version;
}


int Action::SaveList(TagBase *parent, const Vectors *list, Data& data, const Vectors *extra)
{
	Action		*a;
	int		idx, max;
	const Vectors	*saved, *lst;

	saved = extra;

	lst = list;
	while(lst != 0) {
		max = lst->Count();
		for(idx = 0; idx < max; idx++) {
			a = dynamic_cast<Action *>(lst->Get(idx));

//printf("Action[%d] %p: id %d, tag = %p\n", idx, a, a->f_action, a->f_tag);

			// the offset is necessary to compute offsets of branch instructions
			// and also the With operator
			a->f_offset = data.ByteSize();

			// used by the WaitForFrame and DeclareFunction actions
			a->f_index = idx;

			if(a->f_action != ACTION_LABEL) {
				// there is nothing to save in a label
				a->Save(data);
			}
		}
		lst = extra;
		extra = 0;
	}
	data.PutByte(Action::ACTION_END);		// end the list of actions


	/* now we do the 2nd pass! */
	lst = list;
	while(lst != 0) {
		max = lst->Count();
		for(idx = 0; idx < max; idx++) {
			a = dynamic_cast<Action *>(lst->Get(idx));
			a->Save2ndPass(*lst, data);
		}
		lst = saved;
		saved = 0;
	}

	return 0;
}


void Action::Save(Data& data)
{
	Data		sub_data, nested_data;
	Vectors		*sub_list;

	data.PutByte(f_action);

	if(f_action >= 128) {
		sub_list = SubList();
		if(sub_list != 0) {	// Function & With are nested type of entries
			SaveList(Tag(), sub_list, nested_data);
			// get rid of the ACTION_END
			nested_data.Size(nested_data.Size() - CHAR_BIT);

			if(nested_data.ByteSize() >= USHRT_MAX) {
				fprintf(stderr, "ERROR: too many nested instructions; length overflow.\n");
			}
		}

		SaveData(sub_data, nested_data);
		data.PutShort((short) sub_data.ByteSize());
		data.Append(sub_data);

		if(nested_data.ByteSize() > 0) {
			data.Append(nested_data);
		}
	}
}


Vectors *Action::SubList(void)
{
	return 0;
}


int Action::SaveString(Data& data, const char *string)
{
	return f_tag->SaveString(data, string);
}


void Action::SaveData(Data& /*data*/, Data& /*nested_data*/)
{
	assert(0, "the action 0x%02X which has an action number of 128 or more needs to overload the SaveData() function", f_action);
}


void Action::Save2ndPass(const Vectors& list, Data& data)
{
	/* most actions don't require a 2nd pass */
}


Action *Action::Duplicate(void) const
{
	// errors are checked in the object which attach this action to its
	// memory manager
	return new Action(f_tag, f_action);
}


ActionLabel *Action::FindLabel(const Vectors& list, const char *label)
{
	Action		*a;
	ActionLabel	*l;
	int		idx;

	if(label == 0) {
		return 0;
	}

	idx = list.Count();
	while(idx > 0) {
		idx--;
		a = dynamic_cast<Action *>(list.Get(idx));
		if(a->f_action == ACTION_LABEL) {
			l = dynamic_cast<ActionLabel *>(a);
			if(l->GetLabel() != 0) if(strcasecmp(l->GetLabel(), label) == 0) {
				return l;
			}
		}
	}

	return 0;
}



////////////////////////////// Action BRANCH
ActionBranch::ActionBranch(TagBase *tag, action_t action)
	: Action(tag, action)
{
	f_label = 0;
}


void ActionBranch::SetLabel(const char *label)
{
	MemFree(f_label);
	f_label = StrDup(label);
}

Action *ActionBranch::Duplicate(void) const
{
	ActionBranch		*a;

	a = new ActionBranch(Tag(), f_action);
	if(f_label != 0) {
		a->SetLabel(f_label);
	}

	return a;
}


void ActionBranch::SaveData(Data& data, Data& nested_data)
{
	data.PutShort(0);		// unknown offset yet (we do it on the 2nd pass)
}


void ActionBranch::Save2ndPass(const Vectors& list, Data& data)
{
	ActionLabel	*l;
	long		offset;

	// look for that label...
	l = FindLabel(list, f_label);
	if(l == 0) {
		fprintf(stderr, "ERROR: couldn't find label \"%s\" in the list of actions.\n", f_label);
	}
	else {
		// NOTE: the result is a signed value!
		offset = l->Offset() - f_offset - 1 - 2 - 2;
		if(offset < (short) 0x8000 || offset > (short) 0x7FFF) {
			fprintf(stderr, "ERROR: label \"%s\" is out of bounds.\n", f_label);
		}
		else {
			data.OverwriteShort(f_offset + 1 + 2, (short) offset);
		}
	}
}



////////////////////////////// Action CALL FRAME
ActionCallFrame::ActionCallFrame(TagBase *tag)
	: Action(tag, ACTION_CALL_FRAME)
{
}


void ActionCallFrame::SaveData(Data& data, Data& nested_data)
{
	// it is necessary to have a SaveData() function
	// but there isn't any need for us to have anything
	// there since the call frame data is always empty
}


////////////////////////////// Action DICTIONARY
ActionDictionary::ActionDictionary(TagBase *tag)
	: Action(tag, ACTION_DECLARE_DICTIONARY)
{
}


void ActionDictionary::AddString(const char *string)
{
	string_t	*str;

	str = new string_t;
	MemAttach(str, sizeof(string_t), "ActionDictionary::AddString() -- adding a string to the dictionary");
	str->f_string = StrDup(string);
	f_strings.Set(-1, str);
}



Action *ActionDictionary::Duplicate(void) const
{
	string_t		*str;
	ActionDictionary	*d;
	int			max, idx;

	d = new ActionDictionary(Tag());
	max = f_strings.Count();
	for(idx = 0; idx < max; idx++) {
		str = dynamic_cast<string_t *>(f_strings.Get(idx));
		d->AddString(str->f_string);
	}

	return d;
}



void ActionDictionary::SaveData(Data& data, Data& nested_data)
{
	string_t	*str;
	int		max, idx;

	max = f_strings.Count();
	if(max > 256) {
		// we can't save more than 256 strings -- we should
		// warn the user somehow that there is something wrong
		// here!
		max = 256;
	}
	data.PutShort(max);
	for(idx = 0; idx < max; idx++) {
		str = dynamic_cast<string_t *>(f_strings.Get(idx));
		if(SaveString(data, str->f_string) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
	}
}




////////////////////////////// Action FUNCTION
ActionFunction::ActionFunction(TagBase *tag, action_t action)
	: Action(tag, action)
{
	f_name = 0;
	f_registers_count = 0;
	f_flags = 0;
}


void ActionFunction::SetName(const char *name)
{
	MemFree(f_name);
	f_name = StrDup(name);
}


void ActionFunction::SetRegistersCount(unsigned int count)
{
	f_registers_count = count;
}


void ActionFunction::AddParameter(const char *name, int register_number)
{
	parameter_t	*param;

	if(strcmp(name, "this") == 0) {
		f_flags |= ACTION_FUNCTION_LOAD_THIS;
		f_flags &= ~ACTION_FUNCTION_SUPPRESS_THIS;
		return;
	}
	if(strcmp(name, "/this") == 0) {
		f_flags |= ACTION_FUNCTION_SUPPRESS_THIS;
		f_flags &= ~ACTION_FUNCTION_LOAD_THIS;
		return;
	}
	if(strcmp(name, "arguments") == 0) {
		f_flags |= ACTION_FUNCTION_LOAD_ARGUMENTS;
		f_flags &= ~ACTION_FUNCTION_SUPPRESS_ARGUMENTS;
		return;
	}
	if(strcmp(name, "/arguments") == 0) {
		f_flags |= ACTION_FUNCTION_SUPPRESS_ARGUMENTS;
		f_flags &= ~ACTION_FUNCTION_LOAD_ARGUMENTS;
		return;
	}
	if(strcmp(name, "super") == 0) {
		f_flags |= ACTION_FUNCTION_LOAD_SUPER;
		f_flags &= ~ACTION_FUNCTION_SUPPRESS_SUPER;
		return;
	}
	if(strcmp(name, "/super") == 0) {
		f_flags |= ACTION_FUNCTION_SUPPRESS_SUPER;
		f_flags &= ~ACTION_FUNCTION_LOAD_SUPER;
		return;
	}
	if(strcmp(name, "_root") == 0) {
		f_flags |= ACTION_FUNCTION_LOAD_ROOT;
		return;
	}
	if(strcmp(name, "/_root") == 0) {
		f_flags &= ~ACTION_FUNCTION_LOAD_ROOT;
		return;
	}
	if(strcmp(name, "_parent") == 0) {
		f_flags |= ACTION_FUNCTION_LOAD_PARENT;
		return;
	}
	if(strcmp(name, "/_parent") == 0) {
		f_flags &= ~ACTION_FUNCTION_LOAD_PARENT;
		return;
	}
	if(strcmp(name, "_global") == 0) {
		f_flags |= ACTION_FUNCTION_LOAD_GLOBAL;
		return;
	}
	if(strcmp(name, "/_global") == 0) {
		f_flags &= ~ACTION_FUNCTION_LOAD_GLOBAL;
		return;
	}

	assert(register_number > -2 && register_number < 256, "ActionFunction::AddParameter() -- invalid register number specification for an function parameter");

	param = new parameter_t;
	MemAttach(param, sizeof(parameter_t), "ActionFunction::AddParameter() -- parameter_t structure");
	param->f_name = StrDup(name);
	param->f_register_number = register_number;
	f_parameters.Set(-1, param);
}


void ActionFunction::AddAction(Action *action)
{
	f_actions.Set(-1, action);
}


Action *ActionFunction::Duplicate(void) const
{
	parameter_t	*param;
	ActionFunction	*f;
	Action		*a;
	int		max, idx;

	f = new ActionFunction(Tag(), f_action);
	f->SetName(f_name);
	f->SetRegistersCount(f_registers_count);
	f->f_flags = f_flags;	// this wouldn't be set just by duplicating the parameters

	max = f_parameters.Count();
	for(idx = 0; idx < max; idx++) {
		param = dynamic_cast<parameter_t *>(f_parameters.Get(idx));
		f->AddParameter(param->f_name, param->f_register_number);
	}

	max = f_actions.Count();
	for(idx = 0; idx < max; idx++) {
		a = dynamic_cast<Action *>(f_actions.Get(idx));
		f->AddAction(a->Duplicate());
	}

	return f;
}


Vectors *ActionFunction::SubList(void)
{
	return &f_actions;
}


void ActionFunction::SaveData(Data& data, Data& nested_data)
{
	parameter_t	*param;
	int		max, idx, reg;
	char		regs[256];

	if(SaveString(data, f_name) < 0) {
		// TODO: an error occured and this string
		// wasn't saved!
	}

	max = f_parameters.Count();
	data.PutShort(max);

	if(f_action == ACTION_DECLARE_FUNCTION2) {
		data.PutByte(f_registers_count);
		data.PutShort(f_flags);

		// mark the used registers
		memset(regs, 0, sizeof(regs));

		idx = 0;

		if((f_flags & ACTION_FUNCTION_LOAD_THIS) != 0) {
			idx++;
			regs[idx] = 1;
		}
		if((f_flags & ACTION_FUNCTION_LOAD_SUPER) != 0) {
			idx++;
			regs[idx] = 1;
		}
		if((f_flags & ACTION_FUNCTION_LOAD_ROOT) != 0) {
			idx++;
			regs[idx] = 1;
		}
		if((f_flags & ACTION_FUNCTION_LOAD_PARENT) != 0) {
			idx++;
			regs[idx] = 1;
		}
		if((f_flags & ACTION_FUNCTION_LOAD_GLOBAL) != 0) {
			idx++;
			regs[idx] = 1;
		}

		for(idx = 0; idx < max; idx++) {
			param = dynamic_cast<parameter_t *>(f_parameters.Get(idx));
			if(param->f_register_number > 0 && param->f_register_number < 256) {
				regs[param->f_register_number] = 1;
			}
		}

		reg = 1;
		for(idx = 0; idx < max; idx++) {
			param = dynamic_cast<parameter_t *>(f_parameters.Get(idx));
			if(param->f_register_number == 0) {	// auto-assign register?
				// TODO: we need error handling in case more than 256 regs
				// are used (?!?)
				while(reg < 255 && regs[reg] != 0) {
					reg++;
				}
				regs[reg] = 1;
				data.PutByte(reg);
				reg++;
				data.PutByte(0);	// no variable name
			}
			else if(param->f_register_number > 0) {	// auto-assign register?
				// TODO: we need error handling to tell whether the
				// same register is used multiple times
				data.PutByte(param->f_register_number);
				regs[param->f_register_number] = 1;
				data.PutByte(0);	// no variable name
			}
			else {
				data.PutByte(0);	// no register; use variable name
				if(SaveString(data, param->f_name) < 0) {
					// TODO: an error occured and this string
					// wasn't saved!
				}
			}
		}
	}
	else {
		for(idx = 0; idx < max; idx++) {
			param = dynamic_cast<parameter_t *>(f_parameters.Get(idx));
			if(SaveString(data, param->f_name) < 0) {
				// TODO: an error occured and this string
				// wasn't saved!
			}
		}
	}

	// the actions were already saved by the Action::Save() function!
	data.PutShort((unsigned short) nested_data.ByteSize());
}




////////////////////////////// Action GOTO
ActionGoto::ActionGoto(TagBase *tag, action_t action)
	: Action(tag, action)
{
	f_frame_name = 0;
	f_play = false;
}


void ActionGoto::SetFrameName(const char *frame_name)
{
	MemFree(f_frame_name);
	f_frame_name = StrDup(frame_name);
}


Action *ActionGoto::Duplicate(void) const
{
	ActionGoto	*a;

	a = new ActionGoto(Tag(), f_action);
	a->SetFrameName(f_frame_name);
	a->SetPlay(f_play);

	return a;
}


void ActionGoto::SaveData(Data& data, Data& nested_data)
{
	TagBase		*tag;

	switch(f_action) {
	case ACTION_GOTO_LABEL:
		/* in this case the user wants to use a label */
		if(SaveString(data, f_frame_name) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
		break;

	case ACTION_GOTO_FRAME:
		/* here we have to find what ShowFrame is linked to this entry */
		tag = Tag()->FindLabelledTag(f_frame_name);
		if(tag == 0) {
			/* ha! couldn't find this object... */
			fprintf(stderr, "ERROR: can't find any tag labelled '%s'\n", f_frame_name);
			data.PutShort(0);	// a default, just so the file is not boggus.
			return;
		}
		data.PutShort(tag->WhichFrame());
		break;

	case ACTION_GOTO_EXPRESSION:
		data.PutShort(f_play);
		break;

	default:
		assert(0, "invalid f_action for an ActionGoto() object");

	}
}


////////////////////////////// Action LABEL
ActionLabel::ActionLabel(TagBase *tag)
	: Action(tag, ACTION_LABEL)
{
	f_label = 0;
}


void ActionLabel::SetLabel(const char *label)
{
	MemFree(f_label);
	f_label = StrDup(label);
}

Action *ActionLabel::Duplicate(void) const
{
	ActionLabel		*a;

	a = new ActionLabel(Tag());
	if(f_label != 0) {
		a->SetLabel(f_label);
	}

	return a;
}


////////////////////////////// Action PUSH DATA
ActionPushData::ActionPushData(TagBase *tag)
	: Action(tag, ACTION_PUSH_DATA)
{
}


void ActionPushData::AddString(const char *string)
{
	action_immediate_t	*imm;

	imm = new action_immediate_t(ACTION_IMMEDIATE_TYPE_STRING);
	MemAttach(imm, sizeof(action_immediate_t), "ActionPushData::AddString() -- immediate data structure");
	if(string != 0) {
		imm->f_data.f_string = StrDup(string);
	}
	f_data.Set(-1, imm);
}


void ActionPushData::AddBoolean(bool value)
{
	action_immediate_t	*imm;

	imm = new action_immediate_t(ACTION_IMMEDIATE_TYPE_BOOLEAN);
	MemAttach(imm, sizeof(action_immediate_t), "ActionPushData::AddBoolean() -- immediate data structure");
	imm->f_data.f_boolean = value;
	f_data.Set(-1, imm);
}


void ActionPushData::AddInteger(long value)
{
	action_immediate_t	*imm;

	imm = new action_immediate_t(ACTION_IMMEDIATE_TYPE_INTEGER);
	MemAttach(imm, sizeof(action_immediate_t), "ActionPushData::AddInteger() -- immediate data structure");
	imm->f_data.f_integer32 = value;
	f_data.Set(-1, imm);
}


void ActionPushData::AddFloat(float value)
{
	action_immediate_t	*imm;

	imm = new action_immediate_t(ACTION_IMMEDIATE_TYPE_FLOAT);
	MemAttach(imm, sizeof(action_immediate_t), "ActionPushData::AddDouble() -- immediate data structure");
	imm->f_data.f_float32 = value;
	f_data.Set(-1, imm);
}


void ActionPushData::AddDouble(double value)
{
	action_immediate_t	*imm;

	imm = new action_immediate_t(ACTION_IMMEDIATE_TYPE_DOUBLE);
	MemAttach(imm, sizeof(action_immediate_t), "ActionPushData::AddDouble() -- immediate data structure");
	imm->f_data.f_float64 = value;
	f_data.Set(-1, imm);
}


void ActionPushData::AddNull(void)
{
	action_immediate_t	*imm;

	imm = new action_immediate_t(ACTION_IMMEDIATE_TYPE_NULL);
	MemAttach(imm, sizeof(action_immediate_t), "ActionPushData::AddNull() -- immediate data structure");
	f_data.Set(-1, imm);
}


void ActionPushData::AddUndefined(void)
{
	action_immediate_t	*imm;

	imm = new action_immediate_t(ACTION_IMMEDIATE_TYPE_UNDEFINED);
	MemAttach(imm, sizeof(action_immediate_t), "ActionPushData::AddUndefined() -- immediate data structure");
	f_data.Set(-1, imm);
}


void ActionPushData::AddLookup(unsigned short index)
{
	action_immediate_t	*imm;

	imm = new action_immediate_t(index >= 256 ? ACTION_IMMEDIATE_TYPE_LOOKUP_LARGE : ACTION_IMMEDIATE_TYPE_LOOKUP);
	MemAttach(imm, sizeof(action_immediate_t), "ActionPushData::AddLookup() -- immediate data structure");
	imm->f_data.f_lookup = index;
	f_data.Set(-1, imm);
}



void ActionPushData::AddRegister(unsigned char reg_index)
{
	action_immediate_t	*imm;

	assert(reg_index < 4, "invalid register index, limited to the range 0..3 inclusive");

	imm = new action_immediate_t(ACTION_IMMEDIATE_TYPE_REGISTER);
	MemAttach(imm, sizeof(action_immediate_t), "ActionPushData::AddRegister() -- immediate data structure");
	imm->f_data.f_register = reg_index;
	f_data.Set(-1, imm);
}




Action *ActionPushData::Duplicate(void) const
{
	ActionPushData		*a;
	action_immediate_t	*imm;
	int			i, max;

	a = new ActionPushData(Tag());

	max = f_data.Count();
	for(i = 0; i < max; i++) {
		imm = dynamic_cast<action_immediate_t *>(f_data.Get(i));
		switch(imm->f_type) {
		case ACTION_IMMEDIATE_TYPE_STRING:
			a->AddString(imm->f_data.f_string);
			break;

		case ACTION_IMMEDIATE_TYPE_FLOAT:
			a->AddFloat(imm->f_data.f_float32);
			break;

		case ACTION_IMMEDIATE_TYPE_BOOLEAN:
			a->AddBoolean(imm->f_data.f_boolean);
			break;

		case ACTION_IMMEDIATE_TYPE_DOUBLE:
			a->AddDouble(imm->f_data.f_float64);
			break;

		case ACTION_IMMEDIATE_TYPE_INTEGER:
			a->AddInteger(imm->f_data.f_integer32);
			break;

		case ACTION_IMMEDIATE_TYPE_LOOKUP:
		case ACTION_IMMEDIATE_TYPE_LOOKUP_LARGE:
			a->AddLookup(imm->f_data.f_lookup);
			break;

		case ACTION_IMMEDIATE_TYPE_REGISTER:
			a->AddRegister(imm->f_data.f_register);
			break;

		case ACTION_IMMEDIATE_TYPE_NULL:
			a->AddNull();
			break;

		case ACTION_IMMEDIATE_TYPE_UNDEFINED:
			a->AddUndefined();
			break;

		default:
			assert(0, "unknown immediate data type");
			/*NOTREACHED*/

		}
	}

	return a;
}


int ActionPushData::GetMaxRegister(void) const
{
	action_immediate_t	*imm;
	int			register_number;
	int			i;

	register_number = -1;

	i = f_data.Count();
	while(i > 0) {
		i--;
		imm = dynamic_cast<action_immediate_t *>(f_data.Get(i));
		switch(imm->f_type) {
		case ACTION_IMMEDIATE_TYPE_STRING:
		case ACTION_IMMEDIATE_TYPE_FLOAT:
		case ACTION_IMMEDIATE_TYPE_BOOLEAN:
		case ACTION_IMMEDIATE_TYPE_DOUBLE:
		case ACTION_IMMEDIATE_TYPE_INTEGER:
		case ACTION_IMMEDIATE_TYPE_NULL:
		case ACTION_IMMEDIATE_TYPE_UNDEFINED:
			break;

		case ACTION_IMMEDIATE_TYPE_LOOKUP:
		case ACTION_IMMEDIATE_TYPE_LOOKUP_LARGE:
			// here we would need to check in the dictionary
			// to know whether the dictionary is itself defining
			// a register...
			break;

		case ACTION_IMMEDIATE_TYPE_REGISTER:
			if((int) imm->f_data.f_register > register_number) {
				register_number = imm->f_data.f_register;
			}
			break;

		default:
			assert(0, "unknown immediate data type");
			/*NOTREACHED*/

		}
	}

	return register_number;
}


void ActionPushData::SaveData(Data& data, Data& nested_data)
{
	action_immediate_t	*imm;
	int			i, max;

	max = f_data.Count();
	i = 0;
	while(i < max) {
		imm = dynamic_cast<action_immediate_t *>(f_data.Get(i));
		data.PutByte(imm->f_type);
		switch(imm->f_type) {
		case ACTION_IMMEDIATE_TYPE_STRING:
			if(SaveString(data, imm->f_data.f_string) < 0) {
				// TODO: an error occured and this string
				// wasn't saved!
			}
			break;

		case ACTION_IMMEDIATE_TYPE_FLOAT:
			data.PutLong(* (long *) &imm->f_data.f_float32);
			break;

		case ACTION_IMMEDIATE_TYPE_BOOLEAN:
			data.PutByte(imm->f_data.f_boolean);
			break;

		case ACTION_IMMEDIATE_TYPE_DOUBLE:
			data.PutLong(((long *) &imm->f_data.f_float64)[1]);
			data.PutLong(((long *) &imm->f_data.f_float64)[0]);
			break;

		case ACTION_IMMEDIATE_TYPE_INTEGER:
			data.PutLong(imm->f_data.f_integer32);
			break;

		case ACTION_IMMEDIATE_TYPE_REGISTER:
			data.PutByte(imm->f_data.f_register);
			break;

		case ACTION_IMMEDIATE_TYPE_LOOKUP:
			// in this case we know that f_lookup < 256
			data.PutByte((unsigned char) imm->f_data.f_lookup);
			break;

		case ACTION_IMMEDIATE_TYPE_LOOKUP_LARGE:
			data.PutShort(imm->f_data.f_lookup);
			break;

		case ACTION_IMMEDIATE_TYPE_NULL:
		case ACTION_IMMEDIATE_TYPE_UNDEFINED:
			break;

		default:
			assert(0, "unknown immediate data type");
			/*NOTREACHED*/

		}
		i++;
	}
}



////////////////////////////// Action SET TARGET
ActionSetTarget::ActionSetTarget(TagBase *tag)
	: Action(tag, ACTION_SET_TARGET)
{
	f_target = 0;
}


void ActionSetTarget::SetTarget(const char *target)
{
	MemFree(f_target);
	f_target = StrDup(target);
}


Action *ActionSetTarget::Duplicate(void) const
{
	ActionSetTarget	*a;

	a = new ActionSetTarget(Tag());
	a->SetTarget(f_target);

	return a;
}


void ActionSetTarget::SaveData(Data& data, Data& nested_data)
{
	if(SaveString(data, f_target) < 0) {
		// TODO: an error occured and this string
		// wasn't saved!
	}
}





////////////////////////////// Action SET TARGET
ActionStoreRegister::ActionStoreRegister(TagBase *tag)
	: Action(tag, ACTION_STORE_REGISTER)
{
	f_reg = 0;
}


void ActionStoreRegister::SetRegister(unsigned char reg)
{
	f_reg = reg;
}


Action *ActionStoreRegister::Duplicate(void) const
{
	ActionStoreRegister	*a;

	a = new ActionStoreRegister(Tag());
	a->SetRegister(f_reg);

	return a;
}


int ActionStoreRegister::GetMaxRegister(void) const
{
	return f_reg;
}


void ActionStoreRegister::SaveData(Data& data, Data& nested_data)
{
	data.PutByte(f_reg);
}





////////////////////////////// Action TRY
ActionTry::ActionTry(TagBase *tag)
	: Action(tag, ACTION_TRY)
{
	f_register = -1;
	f_variable_name = 0;
	f_has_catch = false;
	f_has_finally = false;
}


void ActionTry::AddTryAction(Action *action)
{
	f_actions_try.Set(-1, action);
}


void ActionTry::AddCatchAction(Action *action)
{
	f_has_catch = true;
	if(action != 0) {
		f_actions_catch.Set(-1, action);
	}
}


void ActionTry::AddFinallyAction(Action *action)
{
	f_has_finally = true;
	if(action != 0) {
		f_actions_finally.Set(-1, action);
	}
}


void ActionTry::SetIdentifier(int register_number)
{
	f_register = register_number;
}


void ActionTry::SetIdentifier(const char *name)
{
	MemFree(f_variable_name);
	f_variable_name = StrDup(name);
}


Action *ActionTry::Duplicate(void) const
{
	ActionTry	*t;
	Action		*a;
	int		max, idx;

	t = new ActionTry(Tag());
	t->SetIdentifier(f_register);
	t->SetIdentifier(f_variable_name);
	t->f_has_catch = f_has_catch;
	t->f_has_finally = f_has_finally;

	max = f_actions_try.Count();
	for(idx = 0; idx < max; idx++) {
		a = dynamic_cast<Action *>(f_actions_try.Get(idx));
		t->AddTryAction(a->Duplicate());
	}

	max = f_actions_catch.Count();
	for(idx = 0; idx < max; idx++) {
		a = dynamic_cast<Action *>(f_actions_catch.Get(idx));
		t->AddCatchAction(a->Duplicate());
	}

	max = f_actions_finally.Count();
	for(idx = 0; idx < max; idx++) {
		a = dynamic_cast<Action *>(f_actions_finally.Get(idx));
		t->AddFinallyAction(a->Duplicate());
	}

	return t;
}


unsigned char ActionTry::Version(void) const
{
	int		v, version;

	// Until version 8, this Version() function is
	// useless... But that's here for forward compatibility
	// since when version 8 comes out I'd probably forget
	// to come back to it!
	version = 7;		// the ActionTry is version 7 by itself

	v = MinimumListVersion(f_actions_try);
	if(v > version) {
		version = v;
	}

	v = MinimumListVersion(f_actions_catch);
	if(v > version) {
		version = v;
	}

	v = MinimumListVersion(f_actions_finally);
	if(v > version) {
		version = v;
	}

	return version;
}


Vectors *ActionTry::SubListTry(void)
{
	return &f_actions_try;
}


Vectors *ActionTry::SubListCatch(void)
{
	f_has_catch = true;
	return &f_actions_catch;
}


Vectors *ActionTry::SubListFinally(void)
{
	f_has_finally = true;
	return &f_actions_finally;
}


void ActionTry::SaveData(Data& data, Data& nested_data)
{
	Data		try_data, catch_data, finally_data;
	unsigned long	try_size, catch_size, finally_size;
	bool		bad;

	assert(f_has_finally || f_has_catch, "an ActionTry needs at least one of CATCH or FINALLY");

	SaveList(Tag(), &f_actions_try, try_data);
	SaveList(Tag(), &f_actions_catch, catch_data);
	SaveList(Tag(), &f_actions_finally, finally_data);

	// get rid of the ACTION_END
	try_data.Size(try_data.Size() - CHAR_BIT);
	catch_data.Size(catch_data.Size() - CHAR_BIT);
	finally_data.Size(finally_data.Size() - CHAR_BIT);

	try_size = try_data.ByteSize();
	catch_size = catch_data.ByteSize();
	finally_size = finally_data.ByteSize();

	if((bad = try_size >= USHRT_MAX)) {
		fprintf(stderr, "ERROR: too many nested instructions in a TRY; length overflow.\n");
	}
	if(catch_size >= USHRT_MAX) {
		fprintf(stderr, "ERROR: too many nested instructions in a CATCH; length overflow.\n");
		bad = true;
	}
	if(finally_size >= USHRT_MAX) {
		fprintf(stderr, "ERROR: too many nested instructions in a FINALLY; length overflow.\n");
		bad = true;
	}
	if(bad) {
		try_size = 0;
		catch_size = 0;
		finally_size = 0;
	}

	data.WriteBits(0, 5);
	data.WriteBits(f_register < 0 ? 0 : 1, 1);
	data.WriteBits(f_has_finally, 1);
	data.WriteBits(f_has_catch, 1);
	data.PutShort((short) finally_size);
	data.PutShort((short) catch_size);
	data.PutShort((short) try_size);
	if(f_register < 0) {
		if(SaveString(data, f_variable_name) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
	}
	else {
		data.PutByte(f_register);
	}
	if(bad) {
		return;
	}

	nested_data.Append(try_data);
	nested_data.Append(catch_data);
	nested_data.Append(finally_data);
}





////////////////////////////// Action URL
ActionURL::ActionURL(TagBase *tag, action_t action)
	: Action(tag, action)
{
	f_url = 0;
	f_target = 0;
	f_method = URL_METHOD_NOVARIABLE;
}


void ActionURL::SetURL(const char *url, const char *target)
{
	MemFree(f_url);
	MemFree(f_target);
	f_url = StrDup(url);
	f_target = StrDup(target);
}


void ActionURL::SetMethod(url_method_t method)
{
	f_method = method;
}


Action *ActionURL::Duplicate(void) const
{
	ActionURL	*a;

	a = new ActionURL(Tag(), f_action);
	a->SetURL(f_url, f_target);
	a->SetMethod(f_method);

	return a;
}


void ActionURL::SaveData(Data& data, Data& nested_data)
{
	if(f_action == ACTION_URL) {
		if(SaveString(data, f_url) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
		if(SaveString(data, f_target) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
	}
	else {
		data.PutByte(f_method);
	}
}






////////////////////////////// Action WAIT FOR FRAME
ActionWaitForFrame::ActionWaitForFrame(TagBase *tag, action_t action)
	: Action(tag, action)
{
	// f_actions = ... -- auto-init with Vectors constructor
	f_frame_name = 0;
}


void ActionWaitForFrame::SetFrameName(const char *frame_name)
{
	MemFree(f_frame_name);
	f_frame_name = StrDup(frame_name);
}


void ActionWaitForFrame::AddAction(Action *action)
{
	f_actions.Set(-1, action);
}


Action *ActionWaitForFrame::Duplicate(void) const
{
	ActionWaitForFrame	*w;
	Action			*a;
	int			idx, max;

	w = new ActionWaitForFrame(Tag(), f_action);
	w->SetFrameName(f_frame_name);
	max = f_actions.Count();
	for(idx = 0; idx < max; idx++) {
		a = dynamic_cast<Action *>(f_actions.Get(idx));
		w->AddAction(a->Duplicate());
	}

	return w;
}


Vectors *ActionWaitForFrame::SubList(void)
{
	return &f_actions;
}


void ActionWaitForFrame::SaveData(Data& data, Data& nested_data)
{
	TagBase		*tag;

	if(f_actions.Count() > 255) {
		fprintf(stderr, "ERROR: too many actions within a WaitForFrame; please, use branches as may be required.\n");
	}

	if(f_action == ACTION_WAIT_FOR_FRAME) {
		if(f_frame_name == 0 || *f_frame_name == '\0') {
			// we save the frame name within the tag in this case
			// so we actually need it!
			fprintf(stderr, "ERROR: a frame name was expected for action WaitForFrame.\n");
			data.PutShort(0);	// a default, just so the file is not boggus.
		}
		else {
			tag = Tag()->FindLabelledTag(f_frame_name);
			if(tag == 0) {
				/* ha! couldn't find this object... */
				fprintf(stderr, "ERROR: can't find any tag labelled '%s'\n", f_frame_name);
				data.PutShort(0);	// a default, just so the file is not boggus.
			}
			else {
				data.PutShort(tag->WhichFrame());
			}
		}
	}

	data.PutByte(f_actions.Count());
}


////////////////////////////// Action WITH
ActionWith::ActionWith(TagBase *tag)
	: Action(tag, ACTION_WITH)
{
}


void ActionWith::AddAction(Action *action)
{
	f_actions.Set(-1, action);
}


Action *ActionWith::Duplicate(void) const
{
	ActionWith	*w;
	Action		*a;
	int		max, idx;

	w = new ActionWith(Tag());
	max = f_actions.Count();
	for(idx = 0; idx < max; idx++) {
		a = dynamic_cast<Action *>(f_actions.Get(idx));
		w->AddAction(a->Duplicate());
	}

	return w;
}


Vectors *ActionWith::SubList(void)
{
	return &f_actions;
}


void ActionWith::SaveData(Data& data, Data& nested_data)
{
	// the actions were already saved by the Action::Save() function!
	data.PutShort((unsigned short) nested_data.ByteSize());
}



