/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;




/////////////////////////////////////////// Edges
Edges::Edges(void)
{
	f_pos = 0;
}


void Edges::Set(int index, const edge_t& edge)
{
	int		idx, max;
	array_edge_t	*e;

	if(index != -1) {
		// verify that this entry exists...
		max = f_edges.Count() * EDGE_BLOCK;
		if(index < max + f_pos) {
			// we have to replace the specified item
			if(index >= max) {
				// that's within the current array
				index -= max;
				f_array.f_edge[index] = edge;
			}
			else {
				// that's in one of the saved block of edges
				idx = index / EDGE_BLOCK;
				index -= idx * EDGE_BLOCK;
				e = dynamic_cast<array_edge_t *>(f_edges.Get(idx));
				e->f_edge[index] = edge;
			}
			return;
		}
		assert(index == max + f_pos, "invalid index of %d in Edges::Set()", index);
		// the index specified means "insert at the end (append)"
	}

	if(f_pos >= EDGE_BLOCK) {
		// not enough room in the current array, save it
		// in our list of vectors to arrays and reset the
		// current array!
		e = new array_edge_t;
		MemAttach(e, sizeof(f_array), "Edges::Set() -- large array");
		idx = EDGE_BLOCK;
		do {
			idx--;
			e->f_edge[idx] = f_array.f_edge[idx];
		} while(idx > 0);
		f_edges.Set(-1, e);
		f_pos = 0;
	}

	f_array.f_edge[f_pos] = edge;
	f_pos++;
}



void Edges::Save(Data& data, long& x, long& y)
{
	int		index, idx, max;
	array_edge_t	*e;

	max = f_edges.Count();
	for(idx = 0; idx < max; idx++) {
		e = dynamic_cast<array_edge_t *>(f_edges.Get(idx));
		for(index = 0; index < EDGE_BLOCK; index++) {
			SaveEdge(data, e->f_edge[index], x, y);
		}
	}

	for(index = 0; index < f_pos; index++) {
		SaveEdge(data, f_array.f_edge[index], x, y);
	}
}



void Edges::SaveEdge(Data& data, const edge_t& edge, long& x, long& y)
{
	int		bits, sz;

	x += edge.f_x;
	y += edge.f_y;

	data.WriteBits(1, 1);		// Edge record
	if(edge.IsLine()) {
		// straight line
		data.WriteBits(1, 1);
		if(edge.f_y == 0.0) {
			// don't need y but we MUST save at least one coordinate so save x here
			// even if it is zero (note that it shouldn't be otherwise these coordinates
			// are really useless!)
			sz = TagBase::SIBitSize(edge.f_x);
			if(sz < 2) {
				sz = 2;
			}
			data.WriteBits(sz - 2, 4);
			data.WriteBits(0, 2);				// only X inserted
			data.WriteBits(edge.f_x, sz);
		}
		else if(edge.f_x == 0.0) {
			// don't need x -- we already know y is necessary
			sz = TagBase::SIBitSize(edge.f_y);
			if(sz < 2) {
				sz = 2;
			}
			data.WriteBits(sz - 2, 4);
			data.WriteBits(1, 2);				// only Y inserted
			data.WriteBits(edge.f_y, sz);
		}
		else {
			// both coordinates necessary
			sz = TagBase::SIBitSize(edge.f_x);
			bits = TagBase::SIBitSize(edge.f_y);
			if(bits > sz) {
				sz = bits;
			}
			if(sz < 2) {
				sz = 2;
			}
			data.WriteBits(sz - 2, 4);
			data.WriteBits(1, 1);
			data.WriteBits(edge.f_x, sz);
			data.WriteBits(edge.f_y, sz);
		}
	}
	else {
		// curve
		x += edge.f_ctrl_x;
		y += edge.f_ctrl_y;

		data.WriteBits(0, 1);
		sz = TagBase::SIBitSize(edge.f_x);
		bits = TagBase::SIBitSize(edge.f_y);
		if(bits > sz) {
			sz = bits;
		}
		bits = TagBase::SIBitSize(edge.f_ctrl_x);
		if(bits > sz) {
			sz = bits;
		}
		bits = TagBase::SIBitSize(edge.f_ctrl_y);
		if(bits > sz) {
			sz = bits;
		}
		if(sz < 2) {
			sz = 2;
		}
		data.WriteBits(sz - 2, 4);
		data.WriteBits(edge.f_ctrl_x, sz);
		data.WriteBits(edge.f_ctrl_y, sz);
		data.WriteBits(edge.f_x, sz);
		data.WriteBits(edge.f_y, sz);
	}
}


