/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;





/////////////////////////////////////////// Event
const Event::event_names_t	Event::f_event_names[SSWF_EVENT_COUNT] = {
	// V5.x
	{ EVENT_ONLOAD,			"ONLOAD" },
	{ EVENT_POINTER_DRAG_ENTER,	"POINTER_DRAG_ENTER" },
	{ EVENT_POINTER_LEAVE,		"POINTER_LEAVE" },
	{ EVENT_POINTER_ENTER,		"POINTER_ENTER" },
	{ EVENT_POINTER_RELEASE_OUTSIDE,"POINTER_RELEASE_OUTSIDE" },
	{ EVENT_POINTER_RELEASE_INSIDE,	"POINTER_RELEASE_INSIDE" },
	{ EVENT_POINTER_PUSH,		"POINTER_PUSH" },
	{ EVENT_INITIALIZE,		"INITIALIZE" },
	{ EVENT_DATA,			"DATA" },

	// V6.x
	{ EVENT_KEY_PRESS,		"KEY_PRESS" },
	{ EVENT_KEY_UP,			"KEY_UP" },
	{ EVENT_KEY_DOWN,		"KEY_DOWN" },
	{ EVENT_POINTER_UP,		"POINTER_UP" },
	{ EVENT_POINTER_DOWN,		"POINTER_DOWN" },
	{ EVENT_POINTER_MOVE,		"POINTER_MOVE" },
	{ EVENT_UNLOAD,			"UNLOAD" },
	{ EVENT_ENTER_FRAME,		"ENTER_FRAME" },
	{ EVENT_POINTER_DRAG_LEAVE,	"DRAG_LEAVE" },

	// Extras for DefineButton2
	{ EVENT_COND_MENU_ENTER,	"MENU_ENTER" },
	{ EVENT_COND_MENU_LEAVE,	"MENU_LEAVE" }
};


const Event::key_names_t	Event::f_key_names[SSWF_KEY_COUNT] = {
	{ KEY_CODE_LEFT_ARROW,	"LEFT_ARROW" },
	{ KEY_CODE_RIGHT_ARROW,	"RIGHT_ARROW" },
	{ KEY_CODE_HOME,	"HOME" },
	{ KEY_CODE_END,		"END" },
	{ KEY_CODE_INSERT,	"INSERT" },
	{ KEY_CODE_DELETE,	"DELETE" },
	{ KEY_CODE_BACKSPACE,	"BACKSPACE" },
	{ KEY_CODE_ENTER,	"ENTER" },
	{ KEY_CODE_UP_ARROW,	"UP_ARROW" },
	{ KEY_CODE_DOWN_ARROW,	"DOWN_ARROW" },
	{ KEY_CODE_PAGE_UP,	"PAGE_UP" },
	{ KEY_CODE_PAGE_DOWN,	"PAGE_DOWN" },
	{ KEY_CODE_TAB,		"TAB" },
	{ KEY_CODE_ESCAPE,	"ESCAPE" },
	{ KEY_CODE_SPACE,	"SPACE" }
};


Event::Event(void)
{
	Reset();
}


void Event::Reset(void)
{
	int		max;

	f_events = 0;

	max = f_actions.Count();
	while(max > 0) {
		max--;
		delete f_actions.Get(max);
	}
	f_actions.Empty();
}


void Event::SetEvents(unsigned long events)
{
	f_events = events;
}


void Event::SetKey(unsigned char key)
{
	f_events |= EVENT_KEY_PRESS;
	f_key = key;
}


unsigned long Event::Events(void) const
{
	return f_events;
}


unsigned long Event::Conditions(void) const
{
	unsigned long	conditions;

	conditions = 0;

	if((f_events & EVENT_KEY_PRESS) != 0) {
		conditions |= f_key << 9;
	}
	if((f_events & EVENT_COND_MENU_LEAVE) != 0) {
		conditions |= CONDITION_MENU_LEAVE;
	}
	if((f_events & EVENT_COND_MENU_ENTER) != 0) {
		conditions |= CONDITION_MENU_ENTER;
	}
	if((f_events & EVENT_POINTER_RELEASE_OUTSIDE) != 0) {
		conditions |= CONDITION_POINTER_RELEASE_OUTSIDE;
	}
	if((f_events & EVENT_POINTER_DRAG_ENTER) != 0) {
		conditions |= CONDITION_POINTER_DRAG_ENTER;
	}
	if((f_events & EVENT_POINTER_DRAG_LEAVE) != 0) {
		conditions |= CONDITION_POINTER_DRAG_LEAVE;
	}
	if((f_events & EVENT_POINTER_RELEASE_INSIDE) != 0) {
		conditions |= CONDITION_POINTER_RELEASE_INSIDE;
	}
	if((f_events & EVENT_POINTER_PUSH) != 0) {
		conditions |= CONDITION_POINTER_PUSH;
	}
	if((f_events & EVENT_POINTER_LEAVE) != 0) {
		conditions |= CONDITION_POINTER_LEAVE;
	}
	if((f_events & EVENT_POINTER_ENTER) != 0) {
		conditions |= CONDITION_POINTER_ENTER;
	}

	return conditions;
}


unsigned char Event::Key(void) const
{
	return f_key;
}


Vectors& Event::Actions(void)
{
	return f_actions;
}


unsigned long Event::StringToEvents(const char *s)
{
	const char	*name;
	char		tmp[32];
	long		idx;
	unsigned long	l, flags;

	flags = 0;

	while(*s != '\0') {
		// clip leading spaces
		while(isspace(*s) || *s == ',') {
			s++;
		}
		name = s;
		while(*s != '\0' && *s != ',') {
			s++;
		}
		// clip ending spaces
		while(s > name && isspace(s[-1])) {
			s--;
		}

		l = s - name;
		assert(l < sizeof(tmp), "event name too long \"%s\"", name);
		if(l < sizeof(tmp)) {
			strncpy(tmp, name, l);
			tmp[l] = '\0';

			if(*tmp >= '0' && *tmp <= '9') {
				flags |= strtoul(tmp, NULL, 0);
			}
			else {
				// search for this name
				for(idx = 0; idx < SSWF_EVENT_COUNT; idx++) {
					if(strcasecmp(tmp, f_event_names[idx].f_name) == 0) {
						flags |= f_event_names[idx].f_flag;
						break;
					}
				}
				// TODO: should we have some kind of error
				//	 reporting here? instead of an assert
				assert(idx < SSWF_EVENT_COUNT, "event \"%s\" unknown", tmp);
			}
		}
	}

	return flags;
}



unsigned char Event::StringToKeyCode(const char *s)
{
	long		idx;

	if(s == 0 || s[0] == '\0') {
		return KEY_CODE_NONE;
	}
	if(s[1] == '\0') {
		// a single character is returned as is (if valid)
		if(s[0] >= 32 && s[0] <= 126) {
			return s[0];
		}
		return KEY_CODE_NONE;
	}

	// search for a corresponding key
	for(idx = 0; idx < SSWF_EVENT_COUNT; idx++) {
		if(strcasecmp(s, f_key_names[idx].f_name) == 0) {
			return f_key_names[idx].f_code;
		}
	}

	return KEY_CODE_NONE;
}


