/* libsswf_memory.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/


#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// MemBuffer
MemBuffer::MemBuffer(void)
{
	f_buffer = 0;
}


MemBuffer::~MemBuffer()
{
	Buffer		*b;

	if(f_buffer != 0) {
		b = f_buffer;
		f_buffer = 0;
		delete b;
	}
}


void MemBuffer::AttachBuffer(Buffer *buffer)
{
	assert(f_buffer == 0 || buffer == 0, "object at 0x%p is being attached to multiple buffers (0x%p [%s] and 0x%p [%s])\n",
				this,
				buffer, buffer == 0 ? "<no info>" : buffer->Info(),
				f_buffer, f_buffer == 0 ? "<no info>" : f_buffer->Info());

	f_buffer = buffer;
}


Buffer *MemBuffer::GetBuffer(void) const
{
	return const_cast<Buffer *>(f_buffer);
}


/////////////////////////////////////////// Buffer
Buffer::Buffer(Buffer **head, size_t size, const char *info)
{
	mem_buffer_t		*data;

	assert(head != 0, "any buffer allocation requires a head buffer pointer");

	f_head = head;
	f_next = *head;
	if(f_next != 0) {
		f_next->f_previous = this;
	}
	f_previous = 0;
	f_info = info;
	f_size = size;

#if DEBUG
	// we adjust the size because some systems may otherwise allocate less bytes
	// and yet our own debug stuff will use this exact size...
	size = SSWF_ALIGN(size, sizeof(long));
#endif

	*head = this;		// link at the head (XXX wouldn't work in multi-thread env. XXX)
	data = (mem_buffer_t *) malloc(size + SSWF_SPACE + SSWF_SAFE * 2);
	if(data == 0) {
		f_data = 0;		// just in case
		fprintf(stderr, "ERROR: out of memory trying to allocate %ld bytes.\n", (long) size);
		assert(0, "check the stack now!");
		exit(1);
	}

	data->f_magic = DMAGIC;
	data->f_buffer = this;
#if DEBUG
	int idx = SSWF_SAFE / sizeof(long);
	long *l = (long *) ((char *) data + SSWF_SPACE);
	long *m = (long *) ((char *) data + SSWF_SPACE + SSWF_SAFE + SSWF_ALIGN(size, sizeof(long)));
	do {
		*l++ = *m++ = SSWF_TEST;
		idx--;
	} while(idx > 0);
#endif
	f_data = (char *) data + SSWF_SPACE + SSWF_SAFE;

#if 0
printf("Allocated buffer at %p (Buffer Object: %p - %s)\n", f_data, this, info);
#endif
}


Buffer::Buffer(Buffer **head, MemBuffer *ptr, size_t size, const char *info)
{
	mem_buffer_t		*data;

	assert(head != 0, "any buffer allocation requires a head buffer pointer");

	f_head = head;
	f_next = *head;
	if(f_next != 0) {
		f_next->f_previous = this;
	}
	f_previous = 0;
	f_info = info;
	f_size = size;
	*head = this;		// link at the head (XXX wouldn't work in multi-thread env. XXX)
		// we have to allocate a SAFE area so the destructor works also
	data = (mem_buffer_t *) malloc(sizeof(MemBuffer *) + SSWF_SPACE + SSWF_SAFE);
	if(data == 0) {
		f_data = 0;		// just in case
		fprintf(stderr, "ERROR: out of memory trying to allocate a memory buffer (MemBuffer) object.\n");
		exit(1);
	}

	data->f_magic = OMAGIC;
	data->f_buffer = this;
	// here we don't initialize the SSWF_SAFE area because it's not really necessary
	f_data = (char *) data + SSWF_SPACE + SSWF_SAFE;
	* (MemBuffer **) f_data = ptr;
	ptr->AttachBuffer(this);

#if 0
printf("Attached pointer %p (Buffer Object: %p - %s)\n", ptr, this, info);
#endif
}


Buffer::~Buffer()
{
	mem_buffer_t	*data;

#if 0
fflush(stdout);
fprintf(stderr, "Free %p - n/p = %p/%p, n.n/p = %p/%p, p.n/p = %p/%p [%s]\n",
		this,
		f_next, f_previous,
		f_next == 0 ? 0 : f_next->f_next, f_next == 0 ? 0 : f_next->f_previous,
		f_previous == 0 ? 0 : f_previous->f_next, f_previous == 0 ? 0 : f_previous->f_previous,
		f_info);
#endif

	if(f_previous == 0) {
		// we are the head, unlinked ourselves & save the new head
		*f_head = f_next;
	}
	else {
		f_previous->f_next = f_next;
	}
	if(f_next != 0) {
		f_next->f_previous = f_previous;
	}

#if DEBUG
	Test();
#endif

	data = (mem_buffer_t *) ((char *) f_data - SSWF_SPACE - SSWF_SAFE);
	if(data->f_magic == OMAGIC) {
		if(((MemBuffer **) f_data)[0]->GetBuffer() != 0) {
#if 0
fflush(stdout);
fprintf(stderr, "Deleting pointer at %p (%s)\n", * (void **) f_data, f_info);
#endif
			((MemBuffer **) f_data)[0]->AttachBuffer(0);		// need to be detached 1st otherwise it will loop
			delete * (MemBuffer **) f_data;
		}
	}
	else if(data->f_magic != DMAGIC) {
		assert(0, "invalid data magic in Buffer::~Buffer() for %p (0x%08lX)", data, data->f_magic);
	}
#if 0
else {
fflush(stdout);
fprintf(stderr, " Freeing buffer  at %p (%s)\n", f_data, f_info);
}
#endif

	free(data);
}




#if DEBUG
void Buffer::Test(void)
{
	bool			bad_start, bad_end;
	int			idx, i;
	long			*l, *m;
	struct mem_buffer_t	*data;

	data = (mem_buffer_t *) ((char *) f_data - SSWF_SPACE - SSWF_SAFE);

//fprintf(stderr, "0x%08lX tested!\n", (long) data);

	if(data->f_magic == DMAGIC) {
		bad_start = false;
		bad_end = false;
		idx = SSWF_SAFE / sizeof(long);
		l = (long *) ((char *) data + SSWF_SPACE);
		m = (long *) ((char *) data + SSWF_SPACE + SSWF_SAFE + SSWF_ALIGN(f_size, sizeof(long)));
		do {
			if(*l++ != SSWF_TEST) {
				bad_start = true;
			}
			if(*m++ != SSWF_TEST) {
				bad_end = true;
			}
			idx--;
		} while(idx > 0);	/* we never expect a bad start/end thus testing the flags here is bad - thanks */
		if(bad_start) {
			idx = SSWF_SAFE / sizeof(long);
			i = 0;
			l = (long *) ((char *) data + SSWF_SPACE);
			do {
				if((i & 7) == 0) {
					fprintf(stderr, "%08X- ", i);
				}
				fprintf(stderr, " %08lX", *l++);
				i++;
				if((i & 7) == 0) {
					fprintf(stderr, "\n");
				}
				idx--;
			} while(idx > 0);
		}
		if(bad_end) {
			idx = SSWF_SAFE / sizeof(long);
			i = SSWF_SAFE;
			m = (long *) ((char *) data + SSWF_SPACE + SSWF_SAFE + SSWF_ALIGN(f_size, sizeof(long)));
			do {
				if((i & 7) == 0) {
					fprintf(stderr, "%08X- ", i);
				}
				fprintf(stderr, " %08lX", *m++);
				i++;
				if((i & 7) == 0) {
					fprintf(stderr, "\n");
				}
				idx--;
			} while(idx > 0);
		}
		if(bad_start || bad_end) {
			assert(0, "invalid memory buffer");
		}
	}
	else if(data->f_magic != OMAGIC) {
		assert(0, "invalid data magic in a data buffer");
	}
}
#endif


void *Buffer::Realloc(size_t size)
{
	mem_buffer_t	*data;

#if DEBUG
	Test();
#endif

	data = (mem_buffer_t *) ((char *) f_data - SSWF_SPACE - SSWF_SAFE);
	assert(data->f_magic == DMAGIC, "can't reallocate a buffer which was not allocated directly (MemAlloc or MemRealloc)");
	data = (mem_buffer_t *) realloc(data, size + SSWF_SPACE + SSWF_SAFE * 2);
	if(data == 0) {
		fprintf(stderr, "ERROR: out of memory reallocating %ld bytes.\n", (long) size);
		exit(1);
	}
#if DEBUG
	int idx = SSWF_SAFE / sizeof(long);
	long *m = (long *) ((char *) data + SSWF_SPACE + SSWF_SAFE + SSWF_ALIGN(size, sizeof(long)));
	do {
		*m++ = SSWF_TEST;
		idx--;
	} while(idx > 0);
#endif

	f_data = (char *) data + SSWF_SPACE + SSWF_SAFE;
	f_size = size;

	return f_data;
}



Buffer *Buffer::FindBuffer(void *ptr)
{
	mem_buffer_t	*data;

	data = (mem_buffer_t *) ((char *) ptr - SSWF_SPACE - SSWF_SAFE);

#ifndef __MINGW32__
	assert(((long) data & (sizeof(double) - 1)) == 0, "invalid buffer alignment at %p", ptr);
#endif
	assert(data->f_magic == DMAGIC || data->f_magic == OMAGIC, "invalid buffer magic word at %p (%08lX)", ptr, data->f_magic);

#if DEBUG
	data->f_buffer->Test();
#endif

	return data->f_buffer;
}



#if 0
// this is actually completely wrong, sorry about that!
bool Buffer::IsBuffer(void *ptr)
{
	mem_buffer_t	*data;

	data = (mem_buffer_t *) ((char *) ptr - SSWF_SPACE - SSWF_SAFE);

	if(((long) data & (sizeof(double) - 1)) != 0) {
		return false;
	}

	return data->f_magic == DMAGIC || data->f_magic == OMAGIC;
}
#endif



/////////////////////////////////////////// Memory Manager

MemoryManager::MemoryManager(void)
{
	f_head = 0;
}


MemoryManager::~MemoryManager()
{
	// delete all of the buffers allocated for this object
	while(f_head != 0) {
		delete f_head;
	}
}


void MemoryManager::MemAttach(MemBuffer *ptr, size_t size, const char *info)
{
	if(ptr == 0) {		// the new ... failed! we catch it here 8-)
		fprintf(stderr, "FATAL ERROR: out of memory allocating an object of %ld bytes.\n", (unsigned long) size);
		exit(1);
	}
	(void) new Buffer(&f_head, ptr, size, info);
}


void *MemoryManager::MemAlloc(size_t size, const char *info)
{
	Buffer		*b;

	b = new Buffer(&f_head, size, info);

	return b->Data();
}


void *MemoryManager::MemRealloc(void *ptr, size_t size, const char *info)
{
	Buffer		*b;

	if(ptr == 0) {
		if(size == 0) {
			return 0;
		}
		return MemAlloc(size, info);
	}

	b = Buffer::FindBuffer(ptr);

	if(size == 0) {
		delete b;
		return 0;
	}

	return b->Realloc(size);
}


unsigned long MemoryManager::Size(void *ptr)
{
	Buffer		*b;

	if(ptr == 0) {
		return 0;
	}

	b = Buffer::FindBuffer(ptr);

	return b->Size();
}


#if DEBUG
void MemoryManager::MemTest(void *ptr)
{
	if(ptr != 0) {
		Buffer::FindBuffer(ptr);
	}
}
#endif


void MemoryManager::MemFree(void *ptr)
{
	if(ptr != 0) {
		delete Buffer::FindBuffer(ptr);
	}
}


void MemoryManager::MemClean(void *ptr)
{
	assert(ptr != 0, "MemClean() can't accept a NULL pointer");

	if(* (void **) ptr != 0) {
		delete Buffer::FindBuffer(* (void **) ptr);
		* (void **) ptr = 0;
	}
}


char *MemoryManager::StrDup(const char *str)
{
	int		l;
	char		*r;

	if(str == 0) {
		l = 0;
	}
	else {
		l = strlen(str);
	}

	r = (char *) MemAlloc(l + 1, "StrDup() -- duplication of a string");

	if(str == 0) {
		r[0] = '\0';
	}
	else {
		strcpy(r, str);
	}

	return r;
}


char *MemoryManager::StrCat(const char *s1, const char *s2)
{
	int		l1, l2;
	char		*r;

	if(s1 == 0) {
		return StrDup(s2);
	}
	if(s2 == 0) {
		return StrDup(s1);
	}

	l1 = strlen(s1);
	l2 = strlen(s2);

	r = (char *) MemAlloc(l1 + l2 + 1, "StrCat() -- concatenate two strings");

	memcpy(r, s1, l1);
	memcpy(r + l1, s2, l2);
	r[l1 + l2] = '\0';

	return r;
}





