/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;

/////////////////////////////////////////// TagDoAction
TagDoAction::TagDoAction(TagBase *parent)
	: TagBase("doaction", parent)
{
	assert(parent != 0, "a Place tag must have a parent");

	f_sprite = 0;
	//f_actions = ... -- auto-init in action!
}



TagBase::swf_type_t TagDoAction::TypeFlags(void) const
{
	return SWF_TYPE_CONTROL | SWF_TYPE_SCRIPT | SWF_TYPE_ONE | SWF_TYPE_REFERENCE;
}


void TagDoAction::SetAction(const Action& action)
{
	Action		*a;

	a = action.Duplicate();
	MemAttach(a, sizeof(Action), "action duplicate in DoAction tag");

	f_actions.Set(-1, a);
}


void TagDoAction::SetSprite(sswf_id_t sprite)
{
	f_sprite = sprite;
}


int TagDoAction::PreSave(void)
{
	if(f_sprite != 0) {
		MinimumVersion(6);
	}

	// at this time, the following isn't necessary if
	// f_sprite != 0, but once v7.x comes out it will
	// be very important to call as well!
	MinimumVersion(Action::MinimumListVersion(f_actions));

	return 0;
}


int TagDoAction::Save(Data& data)
{
	Data		sub_data;

	Action::SaveList(Parent(), &f_actions, sub_data);

	if(f_sprite != 0) {
		SaveTag(data, SWF_TAG_DO_INIT_ACTION, sub_data.ByteSize() + 2);
		data.PutShort(f_sprite);
	}
	else {
		SaveTag(data, SWF_TAG_DO_ACTION, sub_data.ByteSize());
	}
	data.Append(sub_data);

	return 0;
}



