/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;





/////////////////////////////////////////// TagEditText
TagEditText::TagEditText(TagBase *parent)
	: TagBaseID("edit", parent)
{
	//f_bounds = ...
	f_align = EDIT_TEXT_ALIGNMENT_LEFT;
	f_left_margin = 0;
	f_right_margin = 0;
	f_indent = 0;
	f_leading = 0;
	f_font = 0;
	f_font_height = 0;
	f_max_length = 0;
	f_text = 0;
	f_var_name = 0;
	f_used_glyphs = 0;
	f_used_strings = 0;
	//f_color = ...
	f_has_color = false;
	f_word_wrap = true;		// by default most people probably want word wrap (that's all editors default! right?)
	f_multiline = false;
	f_password  = false;
	f_readonly  = false;
	f_no_select = false;
	f_border    = true;		// that's the usual purpose of such an object!
	f_outline   = false;
}


TagBase::swf_type_t TagEditText::TypeFlags(void) const
{
	return SWF_TYPE_DEFINE | SWF_TYPE_REFERENCE;
}


int TagEditText::PreSave(void)
{
	const char		*s;
	char			wname[16];
	TagFont::font_info_t	info;
	TagFont			*font;

	MinimumVersion(f_autosize ? 6 : 4);

	if(f_font == 0 || !f_outline) {
		return 0;
	}

	font = const_cast<TagFont *>(f_font);
	font->SetUsedByEditText();

	// if there are no glyphs, forget the following code
	if(!font->HasGlyph()) {
		return 0;
	}

	// We can't be sure whether the "used strings" below also includes the
	// initialization text so we simply make sure all of the initialization
	// text glyphs are included too.
	if(f_text != 0) {
		s = f_text;
		while(*s != '\0') {
			info.f_glyph = (unsigned char) *s;
			if(!font->FindGlyph(info, true)) {
				// TODO: we need to have an error handler
fprintf(stderr, "ERROR: the character %s doesn't exist in the font named \"%s\". (2)\n", wcname(info.f_glyph, wname), font->FontName());
			}
			s++;
		}
	}

	if(f_used_strings != 0) {
		s = f_used_strings;
		while(*s != '\0') {
			info.f_glyph = (unsigned char) *s;
			if(font->FindGlyph(info, true)) {
				// TODO: we need to have an error handler
fprintf(stderr, "ERROR: the character %s doesn't exist in the font named \"%s\". (3)\n", wcname(info.f_glyph, wname), font->FontName());
			}
			s++;
		}
		// if there is a used string we don't want to call
		// the SetUsedGlyphs() with an empty used glyphs string
		if(f_used_glyphs == 0) {
			return 0;
		}
		if(*f_used_glyphs == '\0') {
			return 0;
		}
	}

	return font->SetUsedGlyphs(f_used_glyphs, true);
}





int TagEditText::Save(Data& data)
{
	Data		sub_data;
	bool		has_layout;

// save this object ID
	SaveID(sub_data);

// save rectangle where text appears
	f_bounds.Save(sub_data);

// save all the flags
	sub_data.Align();
	sub_data.WriteBits(f_text != 0, 1);
	sub_data.WriteBits(f_word_wrap, 1);
	sub_data.WriteBits(f_multiline, 1);
	sub_data.WriteBits(f_password, 1);
	sub_data.WriteBits(f_readonly, 1);
	sub_data.WriteBits(f_has_color, 1);
	sub_data.WriteBits(f_max_length > 0, 1);
	sub_data.WriteBits(f_font != 0, 1);
	sub_data.WriteBits(0, 1);		// reserved
	sub_data.WriteBits(f_autosize, 1);	// auto-size
	has_layout =	   f_align != 0
			|| f_left_margin != 0
			|| f_right_margin != 0
			|| f_indent != 0
			|| f_leading != 0;
	sub_data.WriteBits(has_layout, 1);
	sub_data.WriteBits(f_no_select, 1);
	sub_data.WriteBits(f_border, 1);
	sub_data.WriteBits(0, 1);		// reserved
	sub_data.WriteBits(f_html, 1);

	sub_data.WriteBits(f_outline, 1);
	//sub_data.WriteBits(0, 1);		// considered as reserved - setting to 1 makes the plugins crash!

	if(f_font != 0) {
		f_font->SaveID(sub_data);
		sub_data.PutShort(f_font_height);
	}

	if(f_has_color) {
		f_color.Save(sub_data, true);
	}

	if(f_max_length > 0) {
		sub_data.PutShort((short) f_max_length);
	}

	if(has_layout) {
		sub_data.PutByte((char) f_align);
		sub_data.PutShort((short) f_left_margin);
		sub_data.PutShort((short) f_right_margin);
		sub_data.PutShort((short) f_indent);
		sub_data.PutShort((short) f_leading);
	}

	if(SaveString(sub_data, f_var_name) < 0) {
		// TODO: an error occured and this string
		// wasn't saved!
	}

	if(f_text != 0) {
		if(SaveString(sub_data, f_text) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
	}

// put the tag and size in front of all of that
	SaveTag(data, SWF_TAG_TEXT_FIELD, sub_data.ByteSize());
	data.Append(sub_data);

	return 0;
}


void TagEditText::SetText(const char *text)
{
	MemFree(f_text);
	f_text = StrDup(text);
}


void TagEditText::SetVariableName(const char *name)
{
	MemFree(f_var_name);
	f_var_name = StrDup(name);
}


void TagEditText::SetUsedGlyphs(const char *used_glyphs)
{
	MemFree(f_used_glyphs);
	f_used_glyphs = StrDup(used_glyphs);
}


void TagEditText::AddUsedString(const char *string)
{
	char		*old;

	old = f_used_strings;
	f_used_strings = StrCat(f_used_strings, string);
	MemFree(old);
}


void TagEditText::SetFont(const TagFont *font, long height)
{
	f_font = font;
	f_font_height = (unsigned short) height;
}


void TagEditText::SetMargins(long left, long right)
{
	f_left_margin = left;
	f_right_margin = right;
}


void TagEditText::SetBounds(const SRectangle& bounds)
{
	f_bounds = bounds;
}


void TagEditText::SetAlign(edit_text_alignment_t align)
{
	f_align = align;
}


void TagEditText::SetIndent(long indent)
{
	f_indent = indent;
}


void TagEditText::SetLeading(long leading)
{
	f_leading = leading;
}


void TagEditText::SetColor(Color& color)
{
	f_color = color;
	f_has_color = true;
}


void TagEditText::SetMaxLength(long length)
{
	f_max_length = length;
}


void TagEditText::SetWordWrap(bool status)
{
	f_word_wrap = status;
}


void TagEditText::SetMultiline(bool status)
{
	f_multiline = status;
}


void TagEditText::SetPassword(bool status)
{
	f_password = status;
}


void TagEditText::SetReadOnly(bool status)
{
	f_readonly = status;
}


void TagEditText::SetNoSelect(bool status)
{
	f_no_select = status;
}


void TagEditText::SetBorder(bool status)
{
	f_border = status;
}


void TagEditText::SetOutline(bool status)
{
	f_outline = status;
}


void TagEditText::SetHTML(bool status)
{
	f_html = status;
}


void TagEditText::SetAutoSize(bool status)
{
	f_autosize = status;
}






