/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// TagExport
TagExport::TagExport(TagBase *parent)
	: TagBase("export", parent)
{
	//f_objects = ...
}


TagBase::swf_type_t TagExport::TypeFlags(void) const
{
	return SWF_TYPE_DEFINE | SWF_TYPE_UNIQUE | SWF_TYPE_REFERENCE;
}


void TagExport::SetObject(sswf_id_t id, const char *name)
{
	export_t	*e;

	e = new export_t;
	MemAttach(e, sizeof(export_t), "TagExport::SetObject() -- object info for export");
	e->f_id = id;
	e->f_name = StrDup(name);

	f_objects.Set(-1, e);
}


int TagExport::PreSave(void)
{
	TagBase		*t;
	long		idx, max;
	export_t	*e;

	MinimumVersion(5);

	max = f_objects.Count();
	for(idx = 0; idx < max; idx++) {
		e = dynamic_cast<export_t *>(f_objects.Get(idx));
		t = FindTagWithID(e->f_id, false);
		if(t == 0) {
			// TODO: we need an error handling here
			fprintf(stderr, "ERROR: TagExport: can't find object with ID #%d (given the name \"%s\")\n", e->f_id, e->f_name);
			return -1;
		}
		if((t->TypeFlags() & SWF_TYPE_DEFINE) == 0) {
			// TODO: we need an error handling here
			fprintf(stderr, "ERROR: TagExport: can't export object with ID #%d, it's not a definition tag (given the name \"%s\")\n", e->f_id, e->f_name);
			return -1;
		}
		// if we have a font we want to force it to save all the glyphs
		if(strcmp(t->Name(), "font") == 0) {
			dynamic_cast<TagFont *>(t)->SetUsedByEditText();
		}
	}

	return 0;
}


int TagExport::Save(Data& data)
{
	Data		sub_data;
	long		idx, max;
	export_t	*e;

	max = f_objects.Count();
	sub_data.PutShort((unsigned short) max);

	for(idx = 0; idx < max; idx++) {
		e = dynamic_cast<export_t *>(f_objects.Get(idx));
		sub_data.PutShort(e->f_id);
		if(SaveString(sub_data, e->f_name) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
	}

	MinimumVersion(5);
	SaveTag(data, SWF_TAG_EXPORT, sub_data.ByteSize());

	data.Append(sub_data);

	return 0;
}


