/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// TagFont
TagFont::TagFont(TagBase *parent)
	: TagBaseID("font", parent)
{
	f_font_name = 0;
	f_language = FONT_LANGUAGE_UNKNOWN;
	f_type = FONT_TYPE_BEST;
	f_bold = false;
	f_italic = false;
	f_wide = false;
	f_has_wide_char = false;		// defined in PreSave2ndPass() and used in Save()
	f_has_wide_offsets = false;		// defined in PreSave2ndPass() and used in Save()
	f_used_by_edit_text = false;
	f_define_font2 = false;			// defined in PreSave2ndPass() and used in Save()
	f_has_layout = false;			// defined in PreSave2ndPass() and used in Save()
	f_ascent = DefaultAscent();
	f_descent = DefaultDescent();
	f_leading_height = DefaultLeadingHeight();
	f_default_advance = 0;
	f_space_advance = LONG_MIN;
	f_offsets_max = 0;			// defined in PreSave2ndPass() and used in Save()
	f_offsets = 0;				// defined in PreSave2ndPass() and used in Save()
	f_count = 0;
	//f_save_glyphs = ...
	//f_glyphs = ...
	//f_kerns = ...
}



TagBase::swf_type_t TagFont::TypeFlags(void) const
{
	return SWF_TYPE_DEFINE | SWF_TYPE_HAS_ID;
}


int TagFont::PreSave(void)
{
	long		idx;

	// reset the IN USE flag of each glyph
	idx = f_glyphs.Count();
	while(idx > 0) {
		idx--;
		dynamic_cast<font_glyph_t *>(f_glyphs.Get(idx))->f_in_use = false;
	}

	f_used_by_edit_text = false;

	// if not unknown it was defined by the user
	// if not locale, then we need v6.x
	if(f_language != FONT_LANGUAGE_UNKNOWN
	&& f_language != FONT_LANGUAGE_LOCALE) {
		MinimumVersion(6);
	}

	// a font name with an underscore as the first
	// character is taken as a "direct font name"
	// and thus we need V6.x -- this way we avoid
	// testing specific names and not be forward
	// compatible at a later date when Macromedia
	// decides to add new names.
	if(f_font_name != 0 && *f_font_name == '_') {
		MinimumVersion(6);
	}

	return 0;
}


int TagFont::PreSave2ndPass(void)
{
	font_glyph_t	*glyph;
	int		idx, rmax;
	unsigned long	extra;
	unsigned short	*short_offsets;
#if DEBUG
	long		previous;
#endif

	// Now we know what character is used and whether an
	// edit text is referencing this font - from the
	// Macromedia documents, we have to have a DefineFont2
	// if an edit text references a font
	// Also, if anything said to use v6.x then the fonts
	// must be saved in Unicode (no choice!)
	//
	// WARNING: it is not possible to change from prior
	//	    version 6 to version 6 or more in the
	//	    2nd pass

	f_define_font2 = false;

	// by default we use the user defined f_wide flag
	// which is forced to true if one or more glyph has
	// an ID of 256 or more or we are using v6+
	f_has_wide_char = f_wide || Version() >= 6;

	// to know the size of the glyphs we need to compute it
	// and that means creating the actual buffer of all the
	// glyphs
	rmax = f_glyphs.Count();

	// the default layout flag value is set to true if some
	// of the layout parameters are defined -- note that having
	// a layout in a font doesn't mean we will use them (we do
	// only if a DefineTextField -- an edit text -- references
	// the font)
	f_has_layout = f_ascent != DefaultAscent()
		|| f_descent != DefaultDescent()
		|| f_leading_height != DefaultLeadingHeight()
		|| f_kerns.Count() != 0;
		/* Note: Macromedia says they are not currently taking
		 * the kerning in account... what about other players?
		 */

	// try to avoid re-allocating the offsets buffer if possible
	if(f_offsets != 0) {
		if(f_offsets_max < rmax + 1) {
			MemClean(&f_offsets);
		}
	}
	if(f_offsets == 0) {
		f_offsets = (unsigned long *) MemAlloc((rmax + 1) * sizeof(unsigned long), "offsets to the glyphs");
		f_offsets_max = rmax + 1;
	}
	// by default we have no glyphs (use only the DefineFontInfo[2]
	// if no edit text references this font!)
#if DEBUG
	previous = -1;
#endif
	f_count = 0;
	f_save_glyphs.Empty();
	for(idx = 0; idx < rmax; idx++) {
		glyph = dynamic_cast<font_glyph_t *>(f_glyphs.Get(idx));

#if DEBUG
		// ensure that the glyphs are ordered properly
		assert((long) glyph->f_name > previous, "TagFont::PreSave2ndPass() found unordered glyphs (%ld >= %ld)",
				previous, glyph->f_name);
		previous = glyph->f_name;
#endif

//printf("Glyph [%c] %s\n", glyph->f_name, glyph->f_in_use ? "in use" : "never referenced");

		if(!glyph->f_in_use) {
			// unused, just skip it
			continue;
		}
		if(glyph->f_name > 255) {
			f_has_wide_char = true;
		}
		/*
		 * Note: Macromedia says they don't use the per char
		 *       bounding boxes -- what about other players?
		 */
		if(glyph->f_shape->HasBounds()) {
			f_has_layout = true;
		}
		glyph->f_index = (unsigned short) f_count;		// index at which the glyph is being saved
		f_offsets[f_count] = f_save_glyphs.ByteSize();
		(const_cast<TagShape *>(glyph->f_shape))->SaveWithoutStyles(f_save_glyphs);
		f_save_glyphs.Align();
		f_count++;
	}
	// save the total size in the last offset entry
	f_offsets[f_count] = f_save_glyphs.ByteSize();
	f_count++;
	// got to include this last entry here to make sure we don't
	// overflow anything

	// if an EditText tag references this font then
	// we need a DefineFont2
	// otherwise it depends on whether we need large offsets
	// to access the glyphs
	if(f_used_by_edit_text) {
		// NOTE: the MinimumVersion(3) is useless since the
		//	 TagEditText forces v4.x already!
		//MinimumVersion(3);
		f_define_font2 = true;
	}
	else {
		// we don't want any layout if we don't have
		// an edit text (would be unused!)
		f_has_layout = false;
	}

	// compute the size of the offset table
	extra = (f_count - 1) * 2 + (f_define_font2 ? 2 : 0);
	f_has_wide_offsets = f_offsets[f_count - 1] + extra > USHRT_MAX;
	if(f_has_wide_offsets) {
		MinimumVersion(3);

		// only DefineFont2 accepts large offsets
		if(!f_define_font2) {
			f_define_font2 = true;
			extra += 2;
		}

		// wide offsets! we need to add 2 x extra because
		// that's the actual size of the table! Also include
		// the offset to the map
		extra *= 2;
		for(idx = 0; (unsigned int) idx < f_count; idx++) {
#if BYTE_ORDER == BIG_ENDIAN
			f_offsets[idx] = swap_int(f_offsets[idx] + extra);
#else
			f_offsets[idx] += extra;
#endif
		}
	}
	else {
		short_offsets = reinterpret_cast<unsigned short *>(f_offsets);
		for(idx = 0; (unsigned int) idx < f_count; idx++) {
#if BYTE_ORDER == BIG_ENDIAN
			short_offsets[idx] = swap_short((unsigned short) (f_offsets[idx] + extra));
#else
			short_offsets[idx] = (unsigned short) (f_offsets[idx] + extra);
#endif
		}
	}

	return 0;
}


int TagFont::Save(Data& data)
{
	Data			sub_data, kern_data;
	font_glyph_t		*glyph;
	font_kern_t		*kern;
	int			idx, max, rmax, cnt, used;

	rmax = f_glyphs.Count();

// if the f_font_define2 is set, then we save that type of a
// font, otherwise save a define font + define font info if
// required
	if(f_define_font2) {
		SaveID(sub_data);
		sub_data.WriteBits(f_has_layout, 1);

		// can types be all set? it isn't clear in any doc.
		if(Version() <= 5) {
			switch(f_type) {
			case FONT_TYPE_BEST:
				sub_data.WriteBits(0, 3);		/* not specific */
				break;

			case FONT_TYPE_ASCII:
				sub_data.WriteBits(1, 3);
				break;

			/*
			 * this is now defined as always zero in v6.x doc.
			 * it was properly defined as the Unicode flag
			 * before and is perfectly valid in prior v6.x
			 * movies!
			 */
			case FONT_TYPE_UNICODE:
				sub_data.WriteBits(2, 3);
				break;

			case FONT_TYPE_SHIFTJIS:
				sub_data.WriteBits(4, 3);
				break;

			}
		}
		else {
			sub_data.WriteBits(0, 3);
		}

		// determine whether we need to use wide offsets
		sub_data.WriteBits(f_has_wide_offsets, 1);

		sub_data.WriteBits(f_has_wide_char, 1);

		sub_data.WriteBits(f_italic, 1);
		sub_data.WriteBits(f_bold, 1);

		// language in v6.x, 0 otherwise
		if(Version() <= 5) {
			sub_data.PutByte(0);
		}
		else {
			sub_data.PutByte(f_language);
		}

		cnt = f_font_name == 0 ? 0 : strlen(f_font_name);
		assert(cnt < 256, "a font name can't be more than 255 characters long, change \"%s\"", f_font_name);
		if(cnt > 255) {
			cnt = 255;
		}
		sub_data.PutByte(cnt);		/* this one is a Pstring, why is that, dunno... */
		sub_data.Write(f_font_name, cnt);

		// save the number of glyphs
		// we MUST have one extra "offset" which actually represents the size of the
		// table so one can skip all the glyphs at once
		sub_data.PutShort((short) (f_count - 1));

		// save the array of offsets
		sub_data.Write(f_offsets, f_count * (f_has_wide_offsets ? 4 : 2));

		// save the shapes
		sub_data.Append(f_save_glyphs);

		// save the mapping between the shapes and characters
		for(idx = 0; idx < rmax; idx++) {
			glyph = dynamic_cast<font_glyph_t *>(f_glyphs.Get(idx));
			if(!glyph->f_in_use) {
				// unused, just skip it
				continue;
			}
			if(f_has_wide_char) {
				sub_data.PutShort((unsigned short) glyph->f_name);
			}
			else {
				sub_data.PutByte((unsigned char) glyph->f_name);
			}
		}

		if(f_has_layout) {
			sub_data.PutShort((short) f_ascent);
			sub_data.PutShort((short) f_descent);
			sub_data.PutShort((short) f_leading_height);
			for(idx = 0; idx < rmax; idx++) {
				glyph = dynamic_cast<font_glyph_t *>(f_glyphs.Get(idx));
				if(!glyph->f_in_use) {
					// unused, just skip it
					continue;
				}
				if(glyph->f_advance != LONG_MIN) {
					sub_data.PutShort((short) glyph->f_advance);
				}
				else {
					// if not specific to a character, use the default
					sub_data.PutShort((short) f_default_advance);
				}
			}
			for(idx = 0; idx < rmax; idx++) {
				glyph = dynamic_cast<font_glyph_t *>(f_glyphs.Get(idx));
				if(!glyph->f_in_use) {
					// unused, just skip it
					continue;
				}
				glyph->f_shape->Bounds(0).Save(sub_data);
				sub_data.Align();		// <- is this correct? shouldn't it be in the rectangle Save() function?
			}
			// New max (i.e. the # of kerning entries)
			max = f_kerns.Count();
			for(idx = 0, cnt = 0; idx < max; idx++) {
				kern = dynamic_cast<font_kern_t *>(f_kerns.Get(idx));
				/*
				 * before to save a kern we want to make sure it's not
				 * useless; we go through the list of glyph and ensures
				 * that BOTH the glyphs are defined in the font; if not
				 * then this kerning can't even happen so we don't need
				 * it and thus don't save it.
				 */
				for(idx = 0, used = 0; idx < rmax && used != 3; idx++) {
					glyph = dynamic_cast<font_glyph_t *>(f_glyphs.Get(idx));
					if(glyph->f_name == kern->f_code[0]) {
						if(!glyph->f_in_use) {
							break;
						}
						used |= 1;
					}
					// don't put an 'else' here since code0 == code1
					// is possible! (though probably really rare?)
					if(glyph->f_name == kern->f_code[1]) {
						if(!glyph->f_in_use) {
							break;
						}
						used |= 2;
					}
				}
				if(used == 3) {
					if(f_has_wide_char) {
						kern_data.PutShort((unsigned short) kern->f_code[0]);
						kern_data.PutShort((unsigned short) kern->f_code[1]);
					}
					else {
						kern_data.PutByte((unsigned char) kern->f_code[0]);
						kern_data.PutByte((unsigned char) kern->f_code[1]);
					}
					kern_data.PutShort((short) kern->f_advance);
					cnt++;
				}
			}
			// now we can save the kerning tables
			sub_data.PutShort(cnt);
			sub_data.Append(kern_data);
		}
		SaveTag(data, SWF_TAG_DEFINE_FONT2, sub_data.ByteSize());
		data.Append(sub_data);
	}
	else {
		// if there isn't any glyph we still need a DefineFont tag
		SaveID(sub_data);
		sub_data.Write(f_offsets, f_count == 1 ? 2 : (f_count - 1) * 2);
		sub_data.Append(f_save_glyphs);
		SaveTag(data, SWF_TAG_DEFINE_FONT, sub_data.ByteSize());
		data.Append(sub_data);

		// in some cases we also want to save a DefineFontInfo
		if((f_font_name != 0 && *f_font_name != '\0')
		|| f_bold || f_italic
		|| (f_language != FONT_LANGUAGE_UNKNOWN && f_language != FONT_LANGUAGE_LOCALE)
		|| f_count == 1) {
			sub_data.Empty();
			SaveID(sub_data);	// same ID as the font

			cnt = f_font_name == 0 ? 0 : strlen(f_font_name);
			assert(cnt < 256, "a font name can't be more than 255 characters long, change \"%s\"", f_font_name);
			if(cnt > 255) {
				cnt = 255;
			}
			sub_data.PutByte(cnt);		/* this one is a Pstring, why is that, dunno... */
			sub_data.Write(f_font_name, cnt);

			if(Version() >= 6) {
				sub_data.WriteBits(0, 5);
			}
			else {
				sub_data.WriteBits(0, 2);
				switch(f_type) {
				case FONT_TYPE_BEST:
					sub_data.WriteBits(0, 3);		/* not specific */
					break;

				case FONT_TYPE_ASCII:
					sub_data.WriteBits(1, 3);
					break;

				case FONT_TYPE_SHIFTJIS:
					sub_data.WriteBits(2, 3);
					break;

				/*
				 * this is now defined as always zero in v6.x doc.
				 * it was properly defined as the Unicode flag
				 * before and is perfectly valid in prior v6.x
				 * movies!
				 */
				case FONT_TYPE_UNICODE:
					sub_data.WriteBits(4, 3);
					break;

				}
			}
			sub_data.WriteBits(f_italic, 1);
			sub_data.WriteBits(f_bold, 1);

			// here we know that f_has_wide_char is
			// already 1 when Version() >= 6
			sub_data.WriteBits(f_has_wide_char, 1);

			if(Version() >= 6) {
				sub_data.PutByte(f_language);
			}

			// save the mapping between the shapes and characters
			for(idx = 0; idx < rmax; idx++) {
				glyph = dynamic_cast<font_glyph_t *>(f_glyphs.Get(idx));
				if(!glyph->f_in_use) {
					// unused, just skip it
					continue;
				}
				if(f_has_wide_char) {
					sub_data.PutShort((unsigned short) glyph->f_name);
				}
				else {
					sub_data.PutByte((unsigned char) glyph->f_name);
				}
			}

			SaveTag(data, Version() >= 6 ? SWF_TAG_DEFINE_FONT_INFO2 : SWF_TAG_DEFINE_FONT_INFO, sub_data.ByteSize());
			data.Append(sub_data);
		}
	}

	return 0;
}


const char *	TagFont::f_font_language_name[FONT_LANGUAGE_max] = {
	"LOCALE",
	"LATIN",
	"JAPANESE",
	"KOREAN",
	"SIMPLIFIED_CHINESE",
	"TRADITIONAL_CHINESE"
};


TagFont::font_language_t TagFont::StringToLanguage(const char *language)
{
	font_language_t		idx;

	for(idx = FONT_LANGUAGE_LOCALE; idx < FONT_LANGUAGE_max; idx = (font_language_t) ((int) idx + 1)) {
		if(strcasecmp(f_font_language_name[(int) idx], language) == 0) {
			return idx;
		}
	}

	return FONT_LANGUAGE_UNKNOWN;
}

const char *TagFont::LanguageToString(font_language_t language)
{
	if(language < 0 || language >= FONT_LANGUAGE_max) {
		return "invalid";
	}

	return f_font_language_name[language];
}


void TagFont::AddGlyph(sswf_ucs4_t name, const TagBase *ref, long advance)
{
	font_glyph_t	*glyph;
	font_info_t	info;

	assert(ref != 0, "AddGlyph called without a shape reference");
	assert(strcmp(ref->Name(), "shape") == 0, "a glyph reference must be of type TagShape");

	info.f_glyph = name;
	if(FindGlyph(info)) {
		// the user has the right to redefine the space as an
		// empty character!
		if(info.f_index != SSWF_FONT_SPACE_INDEX) {
			fprintf(stderr, "WARNING: character 'u%ld' defined twice. Second instance ignored\n", name);
			return;
		}
	}

	glyph = new font_glyph_t;
	MemAttach(glyph, sizeof(font_glyph_t), "TagFont::AddGlyph() -- glyph of font");
	glyph->f_name = name;
	glyph->f_shape = dynamic_cast<const TagShape *>(ref);
	glyph->f_advance = advance;	// if LONG_MIN, use the default
	glyph->f_in_use = false;	// if true, must be saved

	// Glyphs MUST be sorted for Macromedia players
	// This allow for binary searches which are very effective
	// and since we searched for the character with the FindGlyph()
	// we already know where we need to insert this new glyph
	f_glyphs.Insert(info.f_position, glyph);
}


void TagFont::AddKern(sswf_ucs4_t code0, sswf_ucs4_t code1, long advance)
{
	font_kern_t	*kern;

	kern = new font_kern_t;
	MemAttach(kern, sizeof(font_kern_t), "TagFont::AddKern() -- font kern");
	kern->f_code[0] = code0;
	kern->f_code[1] = code1;
	kern->f_advance = advance;
	f_kerns.Set(-1, kern);
}


const char *TagFont::FontName(void) const
{
	return f_font_name;
}


/* in this case we expect info.f_index to be properly defined */
void TagFont::GlyphInfo(font_info_t& info) const
{
	font_glyph_t	*glyph;

	assert(info.f_index < (unsigned long) f_glyphs.Count(), "invalid index for a GlyphInfo request");

	glyph = dynamic_cast<font_glyph_t *>(f_glyphs.Get(info.f_index));
	info.f_glyph = glyph->f_name;
	info.f_saved_index = glyph->f_index;
	if(glyph->f_advance == LONG_MIN) {
		info.f_advance = f_default_advance;
	}
	else {
		info.f_advance = glyph->f_advance;
	}
	info.f_is_empty = glyph->f_shape->IsEmpty();
}


bool TagFont::HasGlyph(void) const
{
	return f_glyphs.Count() != 0;
}


/* in this case we expect info.f_glyph to be properly defined */
bool TagFont::FindGlyph(font_info_t& info, bool mark_empty_in_use) const
{
	font_glyph_t	*glyph;
	int		i, j, p;
	bool		found;

	// avoid warnings -- also crashes better if an error occurs
	glyph = 0;

	/*
	 * search the list of available glyphs and
	 * determine the index which represents the character 'c'
	 * note that the glyphs are always ordered
	 */
	p = 0;		// necessary for empty lists!
	found = false;
	j = f_glyphs.Count();
	if(j < 4) {
		/* ugly bubble search when not enough entries yet */
		while(j > 0) {
			j--;
			glyph = dynamic_cast<font_glyph_t *>(f_glyphs.Get(j));
			if(glyph->f_name == info.f_glyph) {
				p = j;
				found = true;
				break;
			}
			if(glyph->f_name < info.f_glyph) {
				// user glyph larger than any of the glyphs
				// in out list
				p = j + 1;
				break;
			}
		}
	}
	else {
		// binary search -- you can't go faster
		i = 0;
		while(i < j) {
			// get the center position of the current range
			p = i + (j - i) / 2;
			glyph = dynamic_cast<font_glyph_t *>(f_glyphs.Get(p));
			if(glyph->f_name == info.f_glyph) {
				found = true;
				break;
			}
			if(glyph->f_name < info.f_glyph) {
				p++;
				i = p;
			}
			else {
				j = p;
			}
		}
	}

	if(found) {
		info.f_index = info.f_position = p;
		info.f_saved_index = glyph->f_index;
		if(glyph->f_advance == LONG_MIN) {
			info.f_advance = f_default_advance;
		}
		else {
			info.f_advance = glyph->f_advance;
		}
		info.f_is_empty = glyph->f_shape->IsEmpty();
		glyph->f_in_use = info.f_is_empty ? mark_empty_in_use : true;
		return true;
	}

	if(info.f_glyph == ' ' && f_space_advance != LONG_MIN) {
		info.f_advance = f_space_advance;
		info.f_index = SSWF_FONT_SPACE_INDEX;
		info.f_position = p;
		info.f_saved_index = 0;		// this one shouldn't be used
		info.f_is_empty = true;
		return true;
	}

	// Ooops doesn't exist!
	// 'p' is the closest position (i.e. if we are adding a new
	// glyph that's where we want it!)
	info.f_index = (unsigned short) (info.f_position = p);
	info.f_saved_index = (unsigned short) info.f_glyph;

	return false;
}



long TagFont::DefaultAscent(void) const
{
	return 750;
}


long TagFont::DefaultDescent(void) const
{
	return 1024 - 750;
}


long TagFont::DefaultLeadingHeight(void) const
{
	return 0;
}


void TagFont::SetName(const char *font_name)
{
	MemFree(f_font_name);
	f_font_name = StrDup(font_name);
}


void TagFont::SetLayout(long ascent, long descent, long leading_height)
{
	f_ascent = ascent;
	f_descent = descent;
	f_leading_height = leading_height;
}


void TagFont::SetDefaultAdvance(long advance)
{
	f_default_advance = advance;
}


void TagFont::SetSpaceAdvance(long advance)
{
	f_space_advance = advance;
}


void TagFont::SetType(font_type_t type)
{
	f_type = type;
}


void TagFont::SetLanguage(font_language_t language)
{
	f_language = language < 0 || language >= FONT_LANGUAGE_max
				? FONT_LANGUAGE_UNKNOWN : language;
}


void TagFont::SetBold(bool bold)
{
	f_bold = bold;
}


void TagFont::SetItalic(bool italic)
{
	f_italic = italic;
}


void TagFont::SetWide(bool wide)
{
	f_wide = wide;
}



int TagFont::SetUsedGlyphs(const char *used_glyphs, bool mark_empty_in_use)
{
	font_info_t	info;
	int		a, b, max, idx;

	// if no glyph, just forget it altogether in this case
	max = f_glyphs.Count();
	if(max == 0) {
		return 0;
	}

	if(used_glyphs == 0 || strcmp(used_glyphs, "*") == 0) {
		// all the characters need to be marked in use
		max = f_glyphs.Count();
		for(idx = 0; idx < max; idx++) {
			dynamic_cast<font_glyph_t *>(f_glyphs.Get(idx))->f_in_use = true;
		}
		return 0;
	}

	while(*used_glyphs != '\0') {
		a = used_glyphs[0];
		if(used_glyphs[1] == '-' && used_glyphs[2] != '\0') {
			b = used_glyphs[2];
			if(a > b) {
				b = a;
				a = used_glyphs[2];
			}
			used_glyphs += 3;
		}
		else {
			b = a;
			used_glyphs++;
		}
		while(a <= b) {
			info.f_glyph = a;
			if(!FindGlyph(info, mark_empty_in_use)) {
				// TODO: we need to have an error handler
fprintf(stderr, "ERROR: the character '%c' (%d) doesn't exist in the font named \"%s\".\n", a, a, f_font_name);
			}
			a++;
		}
	}

	return 0;
}


void TagFont::SetUsedByEditText(bool used)
{
	f_used_by_edit_text = used;
}






