/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;


/////////////////////////////////////////// TagFrameLabel
TagFrameLabel::TagFrameLabel(TagBase *parent)
	: TagBase("framelabel", parent)
{
	f_label = 0;
}




TagBase::swf_type_t TagFrameLabel::TypeFlags(void) const
{
	return SWF_TYPE_CONTROL;
}


void TagFrameLabel::SetFrameLabel(const char *label)
{
	MemFree(f_label);
	f_label = StrDup(label);
}


int TagFrameLabel::PreSave(void)
{
	if(f_label != 0) {
		if(*f_label == '#') {
			MinimumVersion(6);
		}
		else {
			MinimumVersion(3);
		}
	}

	return 0;
}


int TagFrameLabel::Save(Data& data)
{
	long		size, anchor;
	const char	*name;

	name = f_label;
	if(name == 0) {
		return 0;
	}

	size = strlen(name) + 1;

	anchor = *name == '#';

	// NOTE: the size is correct in either case since when we remove
	//	 one character from the name, we have to add 1 char at
	//	 the end of the string

	SaveTag(data, SWF_TAG_FRAME_LABEL, size);

	if(anchor) {
		if(SaveString(data, name + 1) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
		data.PutByte(1);
	}
	else {
		if(SaveString(data, name) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
	}

	return 0;
}


