/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;

/////////////////////////////////////////// TagImport
TagImport::TagImport(TagBase *parent)
	: TagBase("import", parent)
{
	f_url = 0;
	f_filename = 0;
	//f_objects = ...
}


TagBase::swf_type_t TagImport::TypeFlags(void) const
{
	return SWF_TYPE_DEFINE;
}


const char *TagImport::HasID(sswf_id_t id) const
{
	long		idx;
	import_t	*i;

	idx = f_objects.Count();
	while(idx > 0) {
		idx--;
		i = dynamic_cast<import_t *>(f_objects.Get(idx));
		if(i->f_id == id) {
			return i->f_name;
		}
	}

	return 0;
}



sswf_id_t TagImport::HasName(const char *name) const
{
	long		idx;
	import_t	*i;

	idx = f_objects.Count();
	while(idx > 0) {
		idx--;
		i = dynamic_cast<import_t *>(f_objects.Get(idx));
		if(strcmp(i->f_name, name) == 0) {
			return i->f_id;
		}
	}

	return 0;
}



void TagImport::SetURL(const char *url)
{
	MemFree(f_url);
	f_url = StrDup(url);
}


void TagImport::SetFilename(const char *filename)
{
	MemFree(f_filename);
	f_filename = StrDup(filename);
}


void TagImport::AddName(const char *name, const char *type)
{
	import_t	*i;

	i = new import_t;
	MemAttach(i, sizeof(import_t), "TagImport::AddName() -- object info for import");
	i->f_id = Header()->NextID();
	i->f_name = StrDup(name);
	i->f_type = StrDup(type);

	f_objects.Set(-1, i);
}


int TagImport::PreSave(void)
{
	MinimumVersion(5);

	return 0;
}


int TagImport::Save(Data& data)
{
	Data		sub_data;
	long		idx, max;
	import_t	*i;

	if(SaveString(sub_data, f_url) < 0) {
		// TODO: an error occured and this string
		// wasn't saved!
	}

	max = f_objects.Count();
	sub_data.PutShort((unsigned short) max);

	for(idx = 0; idx < max; idx++) {
		i = dynamic_cast<import_t *>(f_objects.Get(idx));
		sub_data.PutShort(i->f_id);
		if(SaveString(sub_data, i->f_name) < 0) {
			// TODO: an error occured and this string
			// wasn't saved!
		}
		// here the type is ignored
	}

	SaveTag(data, SWF_TAG_IMPORT, sub_data.ByteSize());

	data.Append(sub_data);

	return 0;
}


