/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;


/////////////////////////////////////////// TagPlace
TagPlace::TagPlace(TagBase *parent)
	: TagBase("place", parent)
{
	assert(parent != 0, "a Place tag must have a parent");

	f_id_defined = false;
	f_id = 0;		// not defined yet
	f_replace = 1;
	f_depth = -1;
	f_clip_depth = 0;
	f_name = 0;
	f_position = -1;	// not defined yet
	f_has_matrix = false;
	//f_matrix = ...
	//f_color_transform = ...
}


TagBase::swf_type_t TagPlace::TypeFlags(void) const
{
	return SWF_TYPE_CONTROL | SWF_TYPE_SCRIPT | SWF_TYPE_REFERENCE;
}


void TagPlace::SetObjectID(sswf_id_t id)
{
	f_id = id;
	f_id_defined = true;
}


void TagPlace::SetDepth(int depth)
{
	f_depth = depth;
}


void TagPlace::SetClip(int depth)
{
	f_clip_depth = depth;
}


void TagPlace::SetMatrix(const Matrix& matrix)
{
	f_matrix = matrix;
	f_has_matrix = true;
}


void TagPlace::SetColorTransform(const ColorTransform& color_transform)
{
	f_color_transform = color_transform;
}


void TagPlace::SetName(const char *name)
{
	MemFree(f_name);
	f_name = StrDup(name);
}


void TagPlace::SetMorphPosition(int position)
{
	f_position = position;
}


void TagPlace::Replace(void)
{
	f_replace = 0;
}


void TagPlace::Place(void)
{
	f_replace = 1;
}


void TagPlace::AddEvent(Event *event)
{
	f_events.Set(-1, event);
}


int TagPlace::PreSave(void)
{
	int		idx;
	long		v, version;
	Event		*event;

	f_events_all_flags = 0;

	if(f_replace == 0
	|| !f_id_defined
	|| f_position != -1
	|| f_name != 0
	|| !f_color_transform.IsSolidCompatible()
	|| f_clip_depth != 0) {
		/* in this case the order in which we check the events doesn't matter */
		version = 3;
		idx = f_events.Count();
		while(idx > 0) {
			idx--;
			event = dynamic_cast<Event *>(f_events.Get(idx));
			f_events_all_flags |= event->Events();
			v = Action::MinimumListVersion(event->Actions());
			if(v > version) {
				version = v;
			}
		}
		if(f_events_all_flags != 0) {
			if((f_events_all_flags & ~SSWF_EVENT_V5) != 0) {
				// TODO: if any invalid flag is set, return an error
				v = 6;
			}
			else {
				v = 5;
			}
		}
		else {
			v = 3;
		}
		if(v > version) {
			version = v;
		}
		MinimumVersion((unsigned char) v);
	}

	return 0;
}


int TagPlace::Save(Data& data)
{
	Data		sub_data, action_data;
	swf_tag_t	tag;
	int		idx, max;
	unsigned long	flags;
	Event		*event;

	if(f_depth < 0 || f_depth > 65535) {
		// bad depth
		return -1;
	}

	if(f_replace == 0
	|| !f_id_defined
	|| f_position != -1
	|| f_name != 0
	|| !f_color_transform.IsSolidCompatible()
	|| f_clip_depth != 0
	|| f_events_all_flags != 0) {
		tag = SWF_TAG_PLACE_OBJECT2;
		sub_data.WriteBits(f_events_all_flags != 0, 1);	// TODO: events + actions (v5+)
		sub_data.WriteBits(f_clip_depth != 0, 1);
		sub_data.WriteBits(f_name != 0, 1);
		sub_data.WriteBits(f_position != -1, 1);
		sub_data.WriteBits(!f_color_transform.IsNull(true), 1);
		sub_data.WriteBits(f_has_matrix, 1);
		sub_data.WriteBits(f_id_defined, 1);
		/* note that if both f_replace and f_id_defined are zero, then the Macromedia plugin crashes! */
		sub_data.WriteBits(f_replace == 0 && !f_id_defined ? 1 : f_replace, 1);
		sub_data.PutShort(f_depth);
		if(f_id_defined) {
			sub_data.PutShort(f_id);
		}
		if(f_has_matrix) {
			// we have to save null matrices in a PlaceObject2 because we
			// often need the translate of (0, 0)!!!
			f_matrix.Save(sub_data);
		}
		if(!f_color_transform.IsNull(true)) {
			f_color_transform.Save(sub_data, true);
		}
		if(f_position != -1) {
			sub_data.PutShort(f_position);
		}
		if(f_name != 0) {
			if(SaveString(sub_data, f_name) < 0) {
				// TODO: an error occured and this string
				// wasn't saved!
			}
		}
		if(f_clip_depth != 0) {
			sub_data.PutShort((unsigned short) f_clip_depth);
		}
		if(f_events_all_flags != 0) {
			max = f_events.Count();
			if(Version() == 5) {
				/* version 5 uses U16 for event masks (to be confirmed) */
				sub_data.PutShort(0);
				sub_data.PutShort((unsigned short) f_events_all_flags);
				for(idx = 0; idx < max; idx++) {
					event = dynamic_cast<Event *>(f_events.Get(idx));
					sub_data.PutShort((unsigned short) event->Events());
					action_data.Empty();
					Action::SaveList(Parent(), &event->Actions(), action_data);
					sub_data.PutLong(action_data.ByteSize());
					sub_data.Append(action_data);
				}
				sub_data.PutShort(0);
			}
			else {
				/* version 6+ uses U32 for event masks */
				sub_data.PutShort(0);
				sub_data.PutLong(f_events_all_flags);
				for(idx = 0; idx < max; idx++) {
					event = dynamic_cast<Event *>(f_events.Get(idx));
					flags = event->Events();
					sub_data.PutLong(flags);
					action_data.Empty();
					if((flags & Event::EVENT_KEY_PRESS) != 0) {
						action_data.PutByte(event->Key());
					}
					Action::SaveList(Parent(), &event->Actions(), action_data);
					sub_data.PutLong(action_data.ByteSize());
					sub_data.Append(action_data);
				}
				sub_data.PutLong(0);
			}
		}
	}
	else {
		// V1.0 is sufficent for this one
		tag = SWF_TAG_PLACE_OBJECT;
		sub_data.PutShort(f_id);
		sub_data.PutShort(f_depth);
		f_matrix.Save(sub_data);
		if(!f_color_transform.IsNull(false)) {		// save the Color Transformation only if required
			f_color_transform.Save(sub_data, false);
		}
	}

	SaveTag(data, tag, sub_data.ByteSize());
	data.Append(sub_data);

	return 0;
}



