/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// TagProtect
TagProtect::TagProtect(TagBase *parent, swf_tag_t tag)
	: TagBase("protect", parent)
{
	assert(tag == SWF_TAG_PROTECT
		|| tag == SWF_TAG_PROTECT_DEBUG
		|| tag == SWF_TAG_PROTECT_DEBUG2,
			"invalid protection tag, must be one of the SWF_TAG_PROTECT...");

	f_tag = tag;

	if(tag != SWF_TAG_PROTECT) {
		// default "no password" for debug tags
		f_password = "$1$.e$7cXTDev5MooPv3voVnOMX1";
	}
	else {
		f_password = 0;
	}
}



TagBase::swf_type_t TagProtect::TypeFlags(void) const
{
	// here the unicity is a bit particular
	// you can have one PROTECT and
	// one of PROTECT_DEBUG or PROTECT_DEBUG2
	return SWF_TYPE_CONTROL | SWF_TYPE_UNIQUE | SWF_TYPE_START;
}


void TagProtect::SetPassword(const char *password)
{
	// TODO: we need the MD5 encryption right away so we don't have unsafe strings!
	MemFree(f_password);
	f_password = StrDup(password);
	assert(0, "TODO: protect password encryption not supported yet.\n");
}


void TagProtect::SetEncodedPassword(const char *password)
{
	// in this case the caller encoded the password already
	MemFree(f_password);
	f_password = StrDup(password);
}


int TagProtect::PreSave(void)
{
	switch(f_tag) {
	case SWF_TAG_PROTECT:
		MinimumVersion(2);
		break;

	case SWF_TAG_PROTECT_DEBUG:
		MinimumVersion(5);	// TODO: this is actually an ExactVersion(5)... 8-P
		break;			//	 however, we may change this later

	case SWF_TAG_PROTECT_DEBUG2:
		MinimumVersion(6);
		break;

	default:;	// avoid warnings
	/*FALLTHROUGH*/
	}

	return 0;
}


int TagProtect::Save(Data& data)
{
	int		l;
	const char	*password;

	if(f_password == 0 && f_tag == SWF_TAG_PROTECT) {
		// special case where nothing else than the
		// tag is necessary
		SaveTag(data, SWF_TAG_PROTECT, 0);
		return 0;
	}
	if(f_password == 0) {
		password = "";
	}
	else {
		password = f_password;
	}

	l = strlen(password) + 1;

	if(f_tag == SWF_TAG_PROTECT_DEBUG2) {
		SaveTag(data, f_tag, l + 2);
		data.PutShort(0);
	}
	else {
		SaveTag(data, f_tag, l);
	}
	data.Write(f_password, l);

	return 0;
}

