/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



/////////////////////////////////////////// TagSprite
TagSprite::TagSprite(TagBase *parent)
	: TagBaseID("sprite", parent)
{
}



TagBase::swf_type_t TagSprite::TypeFlags(void) const
{
	return SWF_TYPE_DEFINE | SWF_TYPE_REFERENCE | SWF_TYPE_SCRIPT | SWF_TYPE_SPRITE | SWF_TYPE_HAS_ID;
}


int TagSprite::PreSave(void)
{
	MinimumVersion(3);

	return TagBase::PreSave();
}


int TagSprite::Save(Data& data)
{
	TagBase		*child, *last;
	Data		sub_data;
	int		ec;

	// reset the counter in case the user saves multiple times
	ResetFrames();

	SaveID(sub_data);
	sub_data.PutShort(1);		// the frame counter (undefined at this point)

	// save all the children first
	last = 0;
	child = Children();
	while(child != 0) {
		ec = child->Save(sub_data);
		if(ec != 0) {
			return ec;
		}
		last = child;
		child = child->Next();
	}

	// there is nothing in there?!?
	if(last == 0) {
		return 1;
	}

	// add the SHOW FRAME and END tags so it works properly with any player
	// NOTE: if the list ends with and "end" but not "show frame" then there
	//       simply won't be any "show frame".
	if(strcmp(last->Name(), "end") != 0) {
		if(strcmp(last->Name(), "showframe") != 0) {
			ShowFrame();
			SaveTag(sub_data, SWF_TAG_SHOW_FRAME, 0);
		}
		sub_data.PutShort(SWF_TAG_END);
	}

	sub_data.OverwriteShort(sizeof(sswf_id_t), FrameCount());

	SaveTag(data, SWF_TAG_DEFINE_SPRITE, sub_data.ByteSize());
	data.Append(sub_data);

	return 0;
}




