/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;

/////////////////////////////////////////// SRectangle
SRectangle::SRectangle(void)
{
	Reset();
}


void SRectangle::Reset(void)
{
	f_xmin = 0;
	f_xmax = 0;
	f_ymin = 0;
	f_ymax = 0;
}

void SRectangle::Set(long xmin, long xmax, long ymin, long ymax)
{
	f_xmin = xmin;
	f_xmax = xmax;
	f_ymin = ymin;
	f_ymax = ymax;
}


void SRectangle::SetReorder(long xmin, long xmax, long ymin, long ymax)
{
	if(xmin < xmax) {
		f_xmin = xmin;
		f_xmax = xmax;
	}
	else {
		f_xmax = xmin;
		f_xmin = xmax;
	}
	if(ymin < ymax) {
		f_ymin = ymin;
		f_ymax = ymax;
	}
	else {
		f_ymax = ymin;
		f_ymin = ymax;
	}
}

long SRectangle::XMin(void) const
{
	return f_xmin;
}


long SRectangle::XMax(void) const
{
	return f_xmax;
}


long SRectangle::YMin(void) const
{
	return f_ymin;
}


long SRectangle::YMax(void) const
{
	return f_ymax;
}


bool SRectangle::IsEmpty(void) const
{
	return f_xmin >= f_xmax || f_ymin >= f_ymax;
}


void SRectangle::Save(Data& data) const
{
	int		sz, bits;

	sz = TagBase::SIBitSize(f_xmin);
	bits = TagBase::SIBitSize(f_ymin);
	if(bits > sz) {
		sz = bits;
	}
	bits = TagBase::SIBitSize(f_xmax);
	if(bits > sz) {
		sz = bits;
	}
	bits = TagBase::SIBitSize(f_ymax);
	if(bits > sz) {
		sz = bits;
	}

	data.WriteBits(sz, 5);
	data.WriteBits(f_xmin, sz);
	data.WriteBits(f_xmax, sz);
	data.WriteBits(f_ymin, sz);
	data.WriteBits(f_ymax, sz);
}



/////////////////////////////////////////// Color
void Color::Reset(void)
{
	f_red   = 255;
	f_green = 255;
	f_blue  = 255;
	f_alpha = 255;
}


void Color::Set(unsigned char red, unsigned char green, unsigned char blue, unsigned char alpha)
{
	f_red =   red;
	f_green = green;
	f_blue =  blue;
	f_alpha = alpha;
}


void Color::Save(Data& data, bool save_alpha)
{
	data.PutByte(f_red);
	data.PutByte(f_green);
	data.PutByte(f_blue);
	if(save_alpha) {
		data.PutByte(f_alpha);
	}
}


bool Color::operator == (const Color& color) const
{
	return f_red == color.f_red && f_green == color.f_green && f_blue == color.f_blue && f_alpha == color.f_alpha;
}


bool Color::operator != (const Color& color) const
{
	return f_red != color.f_red || f_green != color.f_green || f_blue != color.f_blue || f_alpha != color.f_alpha;
}




/////////////////////////////////////////// ColorTransform
void ColorTransform::Reset(void)
{
	f_add_red   = 0.0;
	f_add_green = 0.0;
	f_add_blue  = 0.0;
	f_add_alpha = 0.0;
	f_mult_red   = 1.0;
	f_mult_green = 1.0;
	f_mult_blue  = 1.0;
	f_mult_alpha = 1.0;
}


void ColorTransform::SetAdd(double red, double green, double blue, double alpha)
{
	f_add_red   = red;
	f_add_green = green;
	f_add_blue  = blue;
	f_add_alpha = alpha;
}


void ColorTransform::SetMult(double red, double green, double blue, double alpha)
{
	f_mult_red   = red;
	f_mult_green = green;
	f_mult_blue  = blue;
	f_mult_alpha = alpha;
}


bool ColorTransform::IsNull(bool with_alpha) const
{
	// if this is true, then this transformation is the Null transformation
	long	ar, ag, ab, aa, mr, mg, mb, ma;

	ar = TagBase::Double2Signed(f_add_red  );
	ag = TagBase::Double2Signed(f_add_green);
	ab = TagBase::Double2Signed(f_add_blue );
	aa = TagBase::Double2Signed(f_add_alpha);

	mr = TagBase::Double2Signed(f_mult_red  );
	mg = TagBase::Double2Signed(f_mult_green);
	mb = TagBase::Double2Signed(f_mult_blue );
	ma = TagBase::Double2Signed(f_mult_alpha);

	if(with_alpha) {
		return ar == 0 && ag == 0 && ab == 0 && aa == 0
			&& mr == 65536 && mg == 65536 && mb == 65536 && ma == 65536;
	}

	return ar == 0 && ag == 0 && ab == 0
			&& mr == 65536 && mg == 65536 && mb == 65536;
}


inline long Double2ColorTransform(double c)
{
	long		r;

	r = (long) rint(c * 256.0);

	/* clamp the result */
	if(r < (short) 0x8000) {
		return (short) 0x8000;
	}
	if(r > (short) 0x7FFF) {
		return (short) 0x7FFF;
	}
	return r;
}

void ColorTransform::Save(Data& data, bool save_alpha)
{
	bool	has_add, has_mult;
	int	sz, bits;
	long	ar, ag, ab, aa, mr, mg, mb, ma;

	ar = Double2ColorTransform(f_add_red  );
	ag = Double2ColorTransform(f_add_green);
	ab = Double2ColorTransform(f_add_blue );
	aa = Double2ColorTransform(f_add_alpha);

	mr = Double2ColorTransform(f_mult_red  );
	mg = Double2ColorTransform(f_mult_green);
	mb = Double2ColorTransform(f_mult_blue );
	ma = Double2ColorTransform(f_mult_alpha);

	if(save_alpha) {
		has_add = ar != 0 || ag != 0 || ab != 0 || aa != 0;
		has_mult = mr != 256 || mg != 256 || mb != 256 || ma != 256;
	}
	else {
		// don't take the alpha in account
		has_add = ar != 0 || ag != 0 || ab != 0;
		has_mult = mr != 256 || mg != 256 || mb != 256;
	}

	sz = 0;

	if(has_add) {
		bits = TagBase::SIBitSize(ar);
		if(bits > sz) {
			sz = bits;
		}
		bits = TagBase::SIBitSize(ag);
		if(bits > sz) {
			sz = bits;
		}
		bits = TagBase::SIBitSize(ab);
		if(bits > sz) {
			sz = bits;
		}
		if(save_alpha) {
			bits = TagBase::SIBitSize(aa);
			if(bits > sz) {
				sz = bits;
			}
		}
	}
	if(has_mult) {
		bits = TagBase::SIBitSize(mr);
		if(bits > sz) {
			sz = bits;
		}
		bits = TagBase::SIBitSize(mg);
		if(bits > sz) {
			sz = bits;
		}
		bits = TagBase::SIBitSize(mb);
		if(bits > sz) {
			sz = bits;
		}
		if(save_alpha) {
			bits = TagBase::SIBitSize(ma);
			if(bits > sz) {
				sz = bits;
			}
		}
	}

	data.Align();
	data.WriteBits(has_add, 1);
	data.WriteBits(has_mult, 1);
	data.WriteBits(sz, 4);

	if(has_mult) {
		data.WriteBits(mr, sz);
		data.WriteBits(mg, sz);
		data.WriteBits(mb, sz);
		if(save_alpha) {
			data.WriteBits(ma, sz);
		}
	}
	if(has_add) {
		data.WriteBits(ar, sz);
		data.WriteBits(ag, sz);
		data.WriteBits(ab, sz);
		if(save_alpha) {
			data.WriteBits(aa, sz);
		}
	}
}





/////////////////////////////////////////// Matrix
Matrix::Matrix(void)
{
	Reset();
}


void Matrix::Reset(void)
{
	f_scale_x = 1.0;
	f_scale_y = 1.0;
	f_rotate = 0.0;
	f_translate_x = 0;
	f_translate_y = 0;
	f_skew_0 = 0.0;
	f_skew_1 = 0.0;
}


void Matrix::SetRotate(double rotate)
{
	// so the comparisons work for similar yet
	// possibly not equal angles
	f_rotate = fmod(rotate, M_PI * 2.0);
}


bool Matrix::IsNull(void) const
{
	return f_scale_x == 1.0 && f_scale_y == 1.0
		&& f_rotate == 0.0
		&& f_translate_x == 0.0 && f_translate_y == 0.0
		&& f_skew_0 == 0.0 && f_skew_1 == 0.0;
}


void Matrix::Save(Data& data)
{
	double		m00, m01, m10, m11, s, c;
	long		x, y, bits, sz;

	s = sin(f_rotate);
	c = cos(f_rotate);
	m00 =  f_scale_x * c;
	m01 =  f_scale_y * s + f_skew_0;
	m10 = -f_scale_x * s + f_skew_1;
	m11 =  f_scale_y * c;

	data.Align();
	if(m00 != 1.0 || m11 != 1.0) {
		data.WriteBits(1, 1);
		x = TagBase::Double2Signed(m00);
		y = TagBase::Double2Signed(m11);
		sz = TagBase::SIBitSize(x);
		bits = TagBase::SIBitSize(y);
		if(bits > sz) {
			sz = bits;
		}
		data.WriteBits(sz, 5);
		data.WriteBits(x, sz);
		data.WriteBits(y, sz);
	}
	else {
		data.WriteBits(0, 1);
	}

	if(m01 != 0.0 || m10 != 0.0) {
		data.WriteBits(1, 1);
		x = TagBase::Double2Signed(m10);
		y = TagBase::Double2Signed(m01);
		sz = TagBase::SIBitSize(x);
		bits = TagBase::SIBitSize(y);
		if(bits > sz) {
			sz = bits;
		}
		data.WriteBits(sz, 5);
		data.WriteBits(x, sz);
		data.WriteBits(y, sz);
	}
	else {
		data.WriteBits(0, 1);
	}

	x = f_translate_x;
	y = f_translate_y;
	sz = TagBase::SIBitSize(x);
	bits = TagBase::SIBitSize(y);
	if(bits > sz) {
		sz = bits;
	}
	if(sz >= 32) {
		fprintf(stderr, "ERROR: invalid matrix - translation too large.\n");
	}
	data.WriteBits(sz, 5);
	if(sz > 0) {
		data.WriteBits(x, sz);
		data.WriteBits(y, sz);
	}
}




bool Matrix::operator == (const Matrix& matrix) const
{
	return f_scale_x == matrix.f_scale_x
		&& f_scale_y == matrix.f_scale_y
		&& f_rotate == matrix.f_rotate
		&& f_translate_x == matrix.f_translate_x
		&& f_translate_y == matrix.f_translate_y;
}


bool Matrix::operator != (const Matrix& matrix) const
{
	return f_scale_x != matrix.f_scale_x
		|| f_scale_y != matrix.f_scale_y
		|| f_rotate != matrix.f_rotate
		|| f_translate_x != matrix.f_translate_x
		|| f_translate_y != matrix.f_translate_y;
}



