/* libsswf_tags.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;



void sswf::swap(void *s1, void *s2, size_t size)
{
	if(size == 0) {
		return;
	}
	if((size & 3) == 0 && (((long) s1) & 3) == 0 && (((long) s2) & 3) == 0) {
		long	a;

		assert(sizeof(long) == 4, "the swap() function assumes that the sizeof(long) == 4...\n");
		do {
			a = * (long *) s1;
			* (long *) s1 = * (long *) s2;
			* (long *) s2 = a;
			s1 = (long *) s1 + 1;
			s2 = (long *) s2 + 1;
			size -= 4;
		} while(size > 0);
	}
	else {
		char	a;

		do {
			a = * (char *) s1;
			* (char *) s1 = * (char *) s2;
			* (char *) s2 = a;
			s1 = (char *) s1 + 1;
			s2 = (char *) s2 + 1;
			size--;
		} while(size > 0);
	}
}


int sswf::wctomb(const sswf_ucs4_t *wc, size_t wc_len, char *mb, size_t& mb_len)
{
	char		buf[6];
	size_t		l;
	sswf_ucs4_t	c;

	while(wc_len >= sizeof(sswf_ucs4_t)) {
		/* get one wide character */
		c = *wc++;
		wc_len -= sizeof(sswf_ucs4_t);
		/* encode in temp buffer */
		if(c < 0x80) {
			/* this will also encode '\0'... */
			buf[0] = (char) c;
			l = 1;
		}
		else if(c < 0x800) {
			buf[0] = (c >> 6) | 0xC0;
			buf[1] = (c & 0x3F) | 0x80;
			l = 2;
		}
		else if(c < 0x10000) {
			buf[0] = (c >> 12) | 0xE0;
			buf[1] = ((c >> 6) & 0x3F) | 0x80;
			buf[2] = (c & 0x3F) | 0x80;
			l = 3;
		}
		else if(c < 0x200000) {
			buf[0] = (c >> 18) | 0xF0;
			buf[1] = ((c >> 12) & 0x3F) | 0x80;
			buf[2] = ((c >> 6) & 0x3F) | 0x80;
			buf[3] = (c & 0x3F) | 0x80;
			l = 4;
		}
		else if(c < 0x4000000) {
			buf[0] = (c >> 24) | 0xF8;
			buf[1] = ((c >> 18) & 0x3F) | 0x80;
			buf[2] = ((c >> 12) & 0x3F) | 0x80;
			buf[3] = ((c >> 6) & 0x3F) | 0x80;
			buf[4] = (c & 0x3F) | 0x80;
			l = 5;
		}
		else if(c > 0) {	// <=> (unsigned long) c < 0x80000000
			buf[0] = (c >> 30) | 0xFC;
			buf[1] = ((c >> 24) & 0x3F) | 0x80;
			buf[2] = ((c >> 18) & 0x3F) | 0x80;
			buf[3] = ((c >> 12) & 0x3F) | 0x80;
			buf[4] = ((c >> 6) & 0x3F) | 0x80;
			buf[5] = (c & 0x3F) | 0x80;
			l = 6;
		}
		else {
			/* an invalid wide character (negative!) */
			return errno = EINVAL;
		}
		if(l < mb_len) {
			mb_len -= l;
			memcpy(mb, buf, l);
			mb += l;
		}
		else {
			/* overflow error */
			return errno = ENOMEM;
		}
	}

	return 0;
}


int sswf::mbtowc(const char *mb, size_t mb_len, sswf_ucs4_t *& wc, size_t& wc_len)
{
	unsigned char	c;
	sswf_ucs4_t	w;
	size_t		l;

	while(mb_len > 0) {
		mb_len--;
		c = (unsigned char) *mb++;
		if(c < 0x80) {
			w = c;
		}
		else {
			if(c >= 0xC0 && c <= 0xDF) {
				l = 1;
				w = c & 0x1F;
			}
			else if(c >= 0xE0 && c <= 0xEF) {
				l = 2;
				w = c & 0x0F;
			}
			else if(c >= 0xF0 && c <= 0xF7) {
				l = 4;
				w = c & 0x07;
			}
			else if(c >= 0xF8 && c <= 0xFB) {
				l = 5;
				w = c & 0x03;
			}
			else if(c >= 0xFC && c <= 0xFD) {
				l = 6;
				w = c & 0x01;
			}
			else {
				return errno = EINVAL;
			}
			if(mb_len < l) {
				return errno = EINVAL;
			}
			mb_len -= l;
			while(l > 0) {
				c = (unsigned char) *mb++;
				if(c < 0x80 || c > 0xBF) {
					return errno = EINVAL;
				}
				l--;
				w = (w << 6) | (c & 0x3F);
			}
		}
		if(wc_len < sizeof(sswf_ucs4_t)) {
			return errno = ENOMEM;
		}
		wc_len -= sizeof(sswf_ucs4_t);
		*wc++ = w;
	}

	return 0;
}


char *sswf::wcname(sswf_ucs4_t wc, char *result)
{
	if(wc >= ' ' && wc <= 0x7E) {
		sprintf(result, "'%c' (%ld)", (char) wc, wc);
	}
	else {
		sprintf(result, "'U+%04lX'", wc);
	}

	return result;
}


long sswf::wclen(sswf_ucs4_t *wcstr)
{
	sswf_ucs4_t	*s;

	if(wcstr == 0) {
		return 0;
	}

	s = wcstr;
	while(*s != '\0') {
		s++;
	}

	return s - wcstr;
}


const char *sswf::sswf_version(void)
{
	static const char *version = TO_STR(SSWF_VERSION);
	return version;
}
