/* libsswf_vectors.c++ -- written by Alexis WILKE for Made to Order Software, Ltd. (c) 2002-2005 */

/*

Copyright (c) 2002-2005 Made to Order Software, Ltd.

Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and
associated documentation files (the "Software"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice
shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include	"sswf/libsswf.h"

using namespace sswf;


/////////////////////////////////////////// Vectors

Vectors::Vectors(void)
{
	f_count = 0;
	f_max = 0;
	f_vectors = 0;
}


Vectors::Vectors(const Vectors& vectors)
{
	f_max = 0;
	f_vectors = 0;
	SetSize(f_count = vectors.f_count);
	if(f_count > 0) {
		// NOTE: the vector data pointer is duplicated but not the objects
		//	 it points to...
		memcpy(f_vectors, vectors.f_vectors, sizeof(ItemBase *) * f_count);
	}
}


Vectors& Vectors::operator = (const Vectors& vectors)
{
	if(this != &vectors) {
		SetSize(f_count = vectors.f_count);
		if(f_count > 0) {
			// NOTE: the vector data pointer is duplicated but not the objects
			//	 it points to...
			memcpy(f_vectors, vectors.f_vectors, sizeof(ItemBase *) * f_count);
		}
	}

	return *this;
}


Vectors::~Vectors()
{
	SetSize(0);		// free everything
}


void Vectors::SetSize(int size)
{
	if(size <= 0) {
		MemClean(&f_vectors);
		f_count = 0;
		f_max = 0;
		return;
	}
	if(size <= f_max) {
		// size already available
		return;
	}
	// round the size up
	size = (size + 255) & -256;

//fprintf(stderr, "Enlarging vector table at %p (%d)\n", this, size);

	f_vectors = (ItemBase **) MemRealloc(f_vectors, size * sizeof(ItemBase *), "Vectors::SetSize() -- create an array of vectors");
	f_max = size;
	if(f_count > size) {
		f_count = size;
	}
}


ItemBase *Vectors::Get(int index) const
{
	assert((unsigned int) index < (unsigned int) f_count, "FATAL ERROR: invalid vector index (%d - max is %d).\n", index, f_count);
	return f_vectors[index];
}





void Vectors::Set(int index, ItemBase *ptr)
{
	if(index == -1) {		// append?
		index = f_count;
	}
	SetSize(index + 1);
	if(index > f_count) {		// ensure all the vectors to that point are cleared
		memset(f_vectors + f_count, 0, (index - f_count) * sizeof(ItemBase *));
	}
	f_vectors[index] = ptr;
	if(f_count < index + 1) {
		f_count = index + 1;
	}
}


void Vectors::Insert(int index, ItemBase *ptr)
{
	if(index == -1) {		// append?
		index = f_count;
	}
	SetSize(index + 1);
	if(index > f_count) {		// ensure all the vectors to that point are cleared
		memset(f_vectors + f_count, 0, (index - f_count) * sizeof(ItemBase *));
	}
	else if(index < f_count) {
		memmove(f_vectors + index + 1, f_vectors + index, (f_count - index) * sizeof(ItemBase *));
	}
	f_vectors[index] = ptr;
	f_count++;
}






